import math
import random
from typing import TypeVar

from pyglet.math import Vec2, Vec3

from arcade.types import AsFloat, HasAddSubMul, Point, Point2, SupportsRichComparison
from arcade.types.rect import Rect
from arcade.types.vector_like import Point3

_PRECISION = 2


__all__ = [
    "clamp",
    "lerp",
    "lerp_2d",
    "lerp_3d",
    "lerp_angle",
    "rand_in_rect",
    "rand_in_circle",
    "rand_on_circle",
    "rand_on_line",
    "rand_angle_360_deg",
    "rand_angle_spread_deg",
    "rand_vec_spread_deg",
    "rand_vec_magnitude",
    "get_distance",
    "rotate_point",
    "get_angle_degrees",
    "get_angle_radians",
    "quaternion_rotation",
]

SupportsRichComparisonT = TypeVar("SupportsRichComparisonT", bound=SupportsRichComparison)


def clamp(
    a: SupportsRichComparisonT, low: SupportsRichComparisonT, high: SupportsRichComparisonT
) -> SupportsRichComparisonT:
    """Clamp a number between a range.

    Args:
        a (float): The number to clamp
        low (float): The lower bound
        high (float): The upper bound
    """
    # Python will deal with > unsupported by falling back on <.
    return high if a > high else max(a, low)  # type: ignore


# This TypeVar helps match v1 and v2 as the same type below in lerp's
# signature. If we used HasAddSubMul, they could be different.
L = TypeVar("L", bound=HasAddSubMul)


def lerp(v1: L, v2: L, u: float) -> L:
    """Linearly interpolate two values which support arithmetic operators.

    Both ``v1`` and ``v2`` must be of compatible types and support
    the following operators:

    * ``+`` (:py:meth:`~object.__add__`)
    * ``-`` (:py:meth:`~object.__sub__`)
    * ``*`` (:py:meth:`~object.__mul__`)

    This means that in certain cases, you may want to use another
    function:

    * For angles, use :py:func:`lerp_angle`.
    * To convert points as arbitary sequences, use:

      * :py:func:`lerp_2d`
      * :py:func:`lerp_3d`

    Args:
        v1 (HasAddSubMul): The first value
        v2 (HasAddSubMul): The second value
        u: The interpolation value `(0.0 to 1.0)`
    """
    return v1 + ((v2 - v1) * u)


def lerp_2d(v1: Point2, v2: Point2, u: float) -> Vec2:
    """Linearly interpolate between two 2D points passed as sequences.

    .. tip:: This function returns a :py:class:`Vec2` you can use
             with :py:func`lerp` .

    Args:
        v1: The first point as a sequence of 2 values.
        v2: The second point as a sequence of 2 values.
        u (float): The interpolation value `(0.0 to 1.0)`
    """
    return Vec2(lerp(v1[0], v2[0], u), lerp(v1[1], v2[1], u))


def lerp_3d(v1: Point3, v2: Point3, u: float) -> Vec3:
    """Linearly interpolate between two 3D points passed as sequences.

    .. tip:: This function returns a :py:class:`Vec3` you can use
             with :py:func`lerp`.

    Args:
        v1: The first point as a sequence of 3 values.
        v2: The second point as a sequence of 3 values.
        u (float): The interpolation value `(0.0 to 1.0)`
    """
    return Vec3(lerp(v1[0], v2[0], u), lerp(v1[1], v2[1], u), lerp(v1[2], v2[2], u))


def smerp(v1: L, v2: L, dt: float, h: float) -> L:
    """
    Smoothly interpolate between two values indepentdant of time.
    use as `a = smerp(a, b, delta_time, 16)`.

    .. tip:: To find the ideal decay constant (half-life) you can use:
             `h = -t / math.log2(p)` where p is how close (percentage) you'd like to be
             to the target value in t seconds.
             i.e if in 1 second you'd like to be within 1% then h ~= 0.15051

    Args:
        v1: The first value to interpolate from.
        v2: The second value to interpolate to.
        dt: The time in seconds that has passed since v1 was interpolated last
        h: The decay constant. The higher the faster v1 reaches v2.
           0.1-25.0 is a good range.
    """

    return v2 + (v1 - v2) * math.pow(2.0, -dt / h)


def smerp_2d(v1: Point2, v2: Point2, dt: float, h: float) -> Vec2:
    """
    Smoothly interpolate between two sequences of length 2 indepentdant of time.
    use as `a = smerp_2d(a, b, delta_time, 16)`.

    .. tip:: To find the ideal decay constant (half-life) you can use:
             `h = -t / math.log2(p)` where p is how close (percentage) you'd like to be
             to the target value in t seconds.
             i.e if in 1 second you'd like to be within 1% then h ~= 0.15051

    .. tip::  This function returns a :py:class:`Vec2` you can use
             with :py:func`smerp`.

    Args:
        v1: The first value to interpolate from.
        v2: The second value to interpolate to.
        dt: The time in seconds that has passed since v1 was interpolated last
        h: The decay constant. The lower the faster v1 reaches v2.
           0.1-25.0 is a good range.
    """
    x1, y1 = v1
    x2, y2 = v2
    d = math.pow(2.0, -dt / h)
    return Vec2(x2 + (x1 - x2) * d, y2 + (y1 - y2) * d)


def smerp_3d(v1: Point3, v2: Point3, dt: float, h: float) -> Vec3:
    """
    Smoothly interpolate between two sequences of length 3 indepentdant of time.
    use as `a = smerp_3d(a, b, delta_time, 16)`.

    .. tip:: To find the ideal decay constant (half-life) you can use:
             `h = -t / math.log2(p)` where p is how close (percentage) you'd like to be
             to the target value in t seconds.
             i.e if in 1 second you'd like to be within 1% then h ~= 0.15051

    .. tip::  This function returns a :py:class:`Vec3` you can use
             with :py:func`smerp`.

    Args:
        v1: The first value to interpolate from.
        v2: The second value to interpolate to.
        dt: The time in seconds that has passed since v1 was interpolated last
        h: The decay constant. The higher the faster v1 reaches v2.
           0.1-25.0 is a good range.
    """
    x1, y1, z1 = v1
    x2, y2, z2 = v2
    d = math.pow(2.0, -dt / h)
    return Vec3(x2 + (x1 - x2) * d, y2 + (y1 - y2) * d, z2 + (z1 - z2) * d)


def lerp_angle(start_angle: float, end_angle: float, u: float) -> float:
    """
    Linearly interpolate between two angles in degrees,
    following the shortest path.

    Args:
        start_angle (float): The starting angle
        end_angle (float): The ending angle
        u (float): The interpolation value (0.0 to 1.0)
    """
    start_angle %= 360
    end_angle %= 360

    while start_angle - end_angle > 180:
        end_angle += 360

    while start_angle - end_angle < -180:
        end_angle -= 360

    return lerp(start_angle, end_angle, u) % 360


def rand_in_rect(rect: Rect) -> Point2:
    """
    Calculate a random point in a rectangle.

    Args:
        rect (Rect): The rectangle to calculate the point in.
    """
    return (
        random.uniform(rect.left, rect.right),
        random.uniform(rect.bottom, rect.top),
    )


def rand_in_circle(center: Point2, radius: float) -> Point2:
    """
    Generate a point in a circle, or can think of it as a vector pointing
    a random direction with a random magnitude <= radius.

    Reference: https://stackoverflow.com/a/50746409

    Args:
        center (Point2): The center of the circle
        radius (float): The radius of the circle
    """
    # random angle
    angle = 2 * math.pi * random.random()
    # random radius
    r = radius * math.sqrt(random.random())
    # calculating coordinates
    return r * math.cos(angle) + center[0], r * math.sin(angle) + center[1]


def rand_on_circle(center: Point2, radius: float) -> Point2:
    """
    Generate a point on a circle.

    Args:
        center (Point2): The center of the circle
        radius (float): The radius of the circle
    """
    angle = 2 * math.pi * random.random()
    return radius * math.cos(angle) + center[0], radius * math.sin(angle) + center[1]


def rand_on_line(pos1: Point2, pos2: Point2) -> Point:
    """
    Given two points defining a line, return a random point on that line.

    Args:
        pos1 (Point2): The first point
        pos2 (Point2): The second point
    """
    u = random.uniform(0.0, 1.0)
    return lerp_2d(pos1, pos2, u)


def rand_angle_360_deg() -> float:
    """
    Returns a random angle in degrees between 0.0 and 360.0.
    """
    return random.uniform(0.0, 360.0)


def rand_angle_spread_deg(angle: float, half_angle_spread: float) -> float:
    """
    Returns a random angle in degrees, within a spread of the given angle.

    Args:
        angle (float): The angle to spread from
        half_angle_spread (float): The half angle spread
    """
    s = random.uniform(-half_angle_spread, half_angle_spread)
    return angle + s


def rand_vec_spread_deg(angle: float, half_angle_spread: float, length: float) -> Point2:
    """
    Returns a random vector, within a spread of the given angle.

    Args:
        angle (float): The angle to spread from
        half_angle_spread (float): The half angle spread
        length (float): The length of the vector
    """
    a = rand_angle_spread_deg(angle, half_angle_spread)
    vel = Vec2.from_polar(a, length)
    return vel.x, vel.y


def rand_vec_magnitude(
    angle: float,
    lo_magnitude: float,
    hi_magnitude: float,
) -> Point2:
    """
    Return a vector of randomized magnitude pointing in the given direction.

    Args:
        angle (float): The vector angle in radians
        lo_magnitude (float): The lower magnitude
        hi_magnitude (float): The higher magnitude
    """
    mag = random.uniform(lo_magnitude, hi_magnitude)
    vel = Vec2.from_polar(angle, mag)
    return vel.x, vel.y


def get_distance(x1: float, y1: float, x2: float, y2: float) -> float:
    """
    Get the distance between two points.

    Args:
        x1 (float): x coordinate of the first point
        y1 (float): y coordinate of the first point
        x2 (float): x coordinate of the second point
        y2 (float): y coordinate of the second point
    """
    return math.hypot(x1 - x2, y1 - y2)


def rotate_point(
    x: float,
    y: float,
    cx: float,
    cy: float,
    angle_degrees: float,
) -> Point2:
    """
    Rotate a point around a center.

    Args:
        x (float): x value of the point you want to rotate
        y (float): y value of the point you want to rotate
        cx (float): x value of the center point you want to rotate around
        cy (float): y value of the center point you want to rotate around
        angle_degrees (float): Angle, in degrees, to rotate
    """
    temp_x = x - cx
    temp_y = y - cy

    # now apply rotation
    angle_radians = math.radians(angle_degrees)
    cos_angle = math.cos(angle_radians)
    sin_angle = math.sin(angle_radians)
    rotated_x = temp_x * cos_angle + temp_y * sin_angle
    rotated_y = -temp_x * sin_angle + temp_y * cos_angle

    # translate back
    x = round(rotated_x + cx, _PRECISION)
    y = round(rotated_y + cy, _PRECISION)

    return x, y


# scale around point


def rescale_relative_to_point(source: Point2, target: Point2, factor: AsFloat | Point2) -> Point2:
    """
    Calculate where a point should be when scaled by the factor realtive to the source point.

    Args:
        source: Where to scaled from.
        target: The point being scaled.
        factor: How much to scale by. If factor is less than one, target approaches source.
                Otherwise it moves away. A factor of zero returns source.

    Returns:
        The rescaled point.
    """

    if isinstance(factor, float | int):
        if factor == 1.0:
            return target
        scale_x = scale_y = factor
    else:
        try:
            scale_x, scale_y = factor
            if scale_x == 1.0 and scale_y == 1.0:
                return target
        except ValueError:
            raise ValueError(
                "factor must be a float, int, or tuple-like which unpacks as two float-like values"
            )
        except TypeError:
            raise TypeError(
                "factor must be a float, int, or tuple-like unpacks as two float-like values"
            )

    dx = target[0] - source[0]
    dy = target[1] - source[1]

    return source[0] + dx * scale_x, source[1] + dy * scale_y


def rotate_around_point(source: Point2, target: Point2, angle: float):
    """
    Rotate a point around another point clockwise.

    Args:
        source: The point to rotate around
        target: The point to rotate
        angle: The degrees to rotate the target by.
    """

    if source == target or angle % 360.0 == 0.0:
        return target

    diff_x = target[0] - source[0]
    diff_y = target[1] - source[1]
    r = math.radians(angle)

    c, s = math.cos(r), math.sin(r)

    dx = diff_x * c - diff_y * s
    dy = diff_x * s + diff_y * c

    return target[0] + dx, target[1] + dy


def get_angle_degrees(x1: float, y1: float, x2: float, y2: float) -> float:
    """
    Get the angle in degrees between two points.

    Args:
        x1 (float): x coordinate of the first point
        y1 (float): y coordinate of the first point
        x2 (float): x coordinate of the second point
        y2 (float): y coordinate of the second point
    """
    x_diff = x2 - x1
    y_diff = y2 - y1
    return -math.degrees(math.atan2(y_diff, x_diff))


def get_angle_radians(x1: float, y1: float, x2: float, y2: float) -> float:
    """
    Get the angle in radians between two points.

    Args:
        x1 (float): x coordinate of the first point
        y1 (float): y coordinate of the first point
        x2 (float): x coordinate of the second point
        y2 (float): y coordinate of the second point
    """
    x_diff = x2 - x1
    y_diff = y2 - y1
    return math.atan2(x_diff, y_diff)


def quaternion_rotation(axis: Point3, vector: Point3, angle: float) -> tuple[float, float, float]:
    """
    Rotate a 3-dimensional vector of any length clockwise around a 3-dimensional unit
    length vector.

    This method of vector rotation is immune to rotation-lock, however it takes a little
    more effort to find the axis of rotation rather than 3 angles of rotation.
    Ref: https://danceswithcode.net/engineeringnotes/quaternions/quaternions.html.

    Args:
        axis (tuple[float, float, float]): The unit length vector that will be rotated around
        vector (tuple[float, float, float]): The 3-dimensional vector to be rotated
        angle (float): The angle in degrees to rotate the vector clock-wise by
    """
    _rotation_rads = -math.radians(angle)
    p1, p2, p3 = vector
    a1, a2, a3 = axis
    _c2, _s2 = math.cos(_rotation_rads / 2.0), math.sin(_rotation_rads / 2.0)

    q0, q1, q2, q3 = _c2, _s2 * a1, _s2 * a2, _s2 * a3
    q0_2, q1_2, q2_2, q3_2 = q0**2, q1**2, q2**2, q3**2
    q01, q02, q03, q12, q13, q23 = q0 * q1, q0 * q2, q0 * q3, q1 * q2, q1 * q3, q2 * q3

    _x = p1 * (q0_2 + q1_2 - q2_2 - q3_2) + 2.0 * (p2 * (q12 - q03) + p3 * (q02 + q13))
    _y = p2 * (q0_2 - q1_2 + q2_2 - q3_2) + 2.0 * (p1 * (q03 + q12) + p3 * (q23 - q01))
    _z = p3 * (q0_2 - q1_2 - q2_2 + q3_2) + 2.0 * (p1 * (q13 - q02) + p2 * (q01 + q23))

    return _x, _y, _z



from PIL import ImageEnhance
from typing_extensions import override

from arcade import Texture
from arcade.gui.events import UIOnChangeEvent, UIOnClickEvent
from arcade.gui.property import Property, bind
from arcade.gui.surface import Surface
from arcade.gui.widgets import UIInteractiveWidget


class UITextureToggle(UIInteractiveWidget):
    """A toggle button switching between on (True) and off (False) state.

    on_texture and off_texture are required.
    State dependent textures are generated by changing the brightness (hover, press)
    of the provided textures or converting them to grayscale (disabled).

    Args:
        x: x coordinate of bottom left
        y: y coordinate of bottom left
        width: Width of the button.
        height: Height of the button.
        on_texture: Texture to show when the button is on.
        off_texture: Texture to show when the button is off.
        value: Initial value of the button.
        size_hint: Size hint for the layout.
        size_hint_min: Minimum size hint for the layout.
        size_hint_max: Maximum size hint for the layout.
    """

    # Experimental ui class
    value = Property(False)

    def __init__(
        self,
        *,
        x: float = 0,
        y: float = 0,
        width: float = 100,
        height: float = 50,
        on_texture: Texture | None = None,
        off_texture: Texture | None = None,
        value=False,
        size_hint=None,
        size_hint_min=None,
        size_hint_max=None,
        **kwargs,
    ):
        # Generate hover and pressed texture by changing the brightness
        if on_texture is None:
            raise ValueError("You have to provide a `on_texture`")
        self.normal_on_tex = on_texture
        enhancer = ImageEnhance.Brightness(self.normal_on_tex.image)
        self.hover_on_tex = Texture(
            enhancer.enhance(1.5),
            name=self.normal_on_tex.cache_name + "_brighter",
        )
        self.pressed_on_tex = Texture(
            enhancer.enhance(0.5),
            name=self.normal_on_tex.cache_name + "_darker",
        )
        self.disabled_on_tex = Texture(
            self.normal_on_tex.image.convert("LA").convert("RGBA"),
        )

        if off_texture is None:
            raise ValueError("You have to provide a `off_texture`")
        self.normal_off_tex = off_texture
        enhancer = ImageEnhance.Brightness(self.normal_off_tex.image)
        self.hover_off_tex = Texture(
            enhancer.enhance(1.5),
            name=self.normal_off_tex.cache_name + "_brighter",
        )
        self.pressed_off_tex = Texture(
            enhancer.enhance(0.5),
            name=self.normal_off_tex.cache_name + "_darker",
        )
        self.disabled_off_tex = Texture(
            self.normal_off_tex.image.convert("LA").convert("RGBA"),
        )

        self.value = value
        self.register_event_type("on_change")

        bind(self, "value", self.trigger_render)
        bind(self, "value", self._dispatch_on_change_event)

        super().__init__(
            x=x,
            y=y,
            width=width,
            height=height,
            size_hint=size_hint,
            size_hint_min=size_hint_min,
            size_hint_max=size_hint_max,
            **kwargs,
        )

    def _dispatch_on_change_event(self):
        self.dispatch_event("on_change", UIOnChangeEvent(self, not self.value, self.value))

    @override
    def on_click(self, event: UIOnClickEvent):
        """Change the value of the button on click."""
        self.value = not self.value

    @override
    def do_render(self, surface: Surface):
        """Render the button, using texture depending on the state."""
        self.prepare_render(surface)
        tex = self.normal_on_tex if self.value else self.normal_off_tex
        if self.disabled:
            tex = self.disabled_on_tex if self.value else self.disabled_off_tex
        elif self.pressed:
            tex = self.pressed_on_tex if self.value else self.pressed_off_tex
        elif self.hovered:
            tex = self.hover_on_tex if self.value else self.hover_off_tex
        surface.draw_texture(0, 0, self.content_width, self.content_height, tex)

    def on_change(self, event: UIOnChangeEvent):
        """To be implemented by the user, triggered when the cursor's value is changed.

        Args:
            event: Event containing the old and new value of the cursor.
        """
        pass



"""
Post-processing shaders.
"""

from arcade import get_window
from arcade.context import ArcadeContext
from arcade.experimental.gaussian_kernel import gaussian_kernel
from arcade.gl import geometry
from arcade.gl.texture import Texture2D


class PostProcessing:
    """Base class"""

    def __init__(self, size: tuple[int, int], *args, **kwargs):
        self._size = size
        window = get_window()
        if not window:
            raise RuntimeError("No window found")
        self._ctx = window.ctx

    @property
    def ctx(self) -> ArcadeContext:
        """Get the shader context."""
        return self._ctx

    @property
    def width(self):
        """Get the width of the buffer."""
        return self._size[0]

    @property
    def height(self):
        """Get the height of the buffer."""
        return self._size[1]

    @property
    def size(self):
        """Get the size of the buffer."""
        return self._size

    # def render(self, *args, **kwargs) -> Texture:
    #     """ Render. Should be over-loaded by the child class. """
    #     pass

    def resize(self, size: tuple[int, int]):
        """
        Resize post processing buffers.
        This is often needed of the screen since changes.
        """
        self._size = size


class GaussianBlurPass(PostProcessing):
    def __init__(self, size, kernel_size=5, sigma=2, multiplier=0, step=1):
        super().__init__(size)
        self._kernel_size = kernel_size
        self._sigma = sigma
        self._multiplier = multiplier
        self._step = step

    def _create_kernel(self):
        """Create kernel and return defines for shader"""
        kernel = gaussian_kernel(self._kernel_size, self._sigma, self._multiplier, self._step)
        if self._multiplier != 1:
            for i in range(self._kernel_size):
                kernel[i] *= self._multiplier

        kernel_string = "("
        for index, item in enumerate(kernel):
            kernel_string += f"{item:.7}"
            if index < len(kernel) - 1:
                kernel_string += ", "
        kernel_string += ")"

        return {"KERNEL_SIZE": str(self._kernel_size), "MY_KERNEL": kernel_string}


class GaussianBlurHorizontal(GaussianBlurPass):
    """Blur the buffer horizontally."""

    def __init__(self, size: tuple[int, int], kernel_size=5, sigma=2, multiplier=1, step=1):
        super().__init__(size, kernel_size=5, sigma=2, multiplier=1, step=1)
        color_attachment = self.ctx.texture(
            size, components=3, wrap_x=self.ctx.CLAMP_TO_EDGE, wrap_y=self.ctx.CLAMP_TO_EDGE
        )
        self._fbo = self.ctx.framebuffer(color_attachments=color_attachment)
        self._program = self.ctx.load_program(
            # defines=self._create_kernel(),
            vertex_shader=":system:shaders/texture_default_projection_vs.glsl",
            fragment_shader=":system:shaders/postprocessing/gaussian_blur_x_fs.glsl",
        )
        self._quad_fs = geometry.quad_2d_fs()

    def render(self, source: Texture2D) -> Texture2D:
        """Render"""
        self._fbo.use()
        source.use(0)
        self._program["target_size"] = self._fbo.size
        self._quad_fs.render(self._program)
        return self._fbo.color_attachments[0]


class GaussianBlurVertical(GaussianBlurPass):
    """Blur the buffer vertically."""

    def __init__(self, size: tuple[int, int], kernel_size=5, sigma=2, multiplier=1, step=1):
        super().__init__(
            size, kernel_size=kernel_size, sigma=sigma, multiplier=multiplier, step=step
        )
        self._fbo = self.ctx.framebuffer(
            color_attachments=self.ctx.texture(
                size,
                components=3,
                wrap_x=self.ctx.CLAMP_TO_EDGE,
                wrap_y=self.ctx.CLAMP_TO_EDGE,
            )
        )
        self._program = self.ctx.load_program(
            # defines=self._create_kernel(),
            vertex_shader=":system:shaders/texture_default_projection_vs.glsl",
            fragment_shader=":system:shaders/postprocessing/gaussian_blur_y_fs.glsl",
        )
        self._quad_fs = geometry.quad_2d_fs()

    def render(self, source: Texture2D) -> Texture2D:
        """Render"""
        self._fbo.use()
        source.use(0)
        self._program["target_size"] = self._fbo.size
        self._quad_fs.render(self._program)
        return self._fbo.color_attachments[0]


class GaussianBlur(PostProcessing):
    """Do both horizontal and vertical blurs."""

    def __init__(self, size, kernel_size=5, sigma=2, multiplier=0, step=1):
        super().__init__(size)
        self._blur_x = GaussianBlurHorizontal(
            size, kernel_size=kernel_size, sigma=sigma, multiplier=multiplier, step=step
        )
        self._blur_y = GaussianBlurVertical(
            size, kernel_size=kernel_size, sigma=sigma, multiplier=multiplier, step=step
        )

    def render(self, source: Texture2D) -> Texture2D:
        """Render"""
        blurred_x = self._blur_x.render(source)
        return self._blur_y.render(blurred_x)


class BloomEffect(PostProcessing):
    """Post processing to create a bloom/glow effect."""

    def __init__(
        self,
        size,
        kernel_size=5,
        sigma: float = 2,
        mu: float = 0,
        multiplier: float = 5,
        step: int = 1,
    ):
        super().__init__(size, kernel_size=kernel_size)

        self._gaussian_1 = GaussianBlur(
            (size[0], size[1]),
            kernel_size=kernel_size,
            sigma=sigma,
            multiplier=multiplier,
            step=step,
        )
        self._gaussian_2 = GaussianBlur(
            (size[0] // 4, size[1] // 4),
            kernel_size=kernel_size,
            sigma=sigma,
            multiplier=multiplier,
            step=step,
        )
        self._gaussian_3 = GaussianBlur(
            (size[0] // 8, size[1] // 8),
            kernel_size=kernel_size,
            sigma=sigma,
            multiplier=multiplier,
            step=step,
        )

        # Program and buffer doing contrast / brightness / luma
        luma_tex = self.ctx.texture((self.width // 2, self.height // 2), components=3)
        luma_tex.wrap_x = self.ctx.CLAMP_TO_EDGE
        luma_tex.wrap_y = self.ctx.CLAMP_TO_EDGE
        self._cb_luma_buffer = self.ctx.framebuffer(color_attachments=[luma_tex])
        # Buffer for the converted luma values
        self._cb_luma_program = self.ctx.load_program(
            vertex_shader=":system:shaders/postprocessing/glow_filter_vs.glsl",
            fragment_shader=":system:shaders/postprocessing/glow_filter_fs.glsl",
        )

        # Program for combining the original buffer and the blurred buffer
        self._combine_program = self.ctx.load_program(
            vertex_shader=":system:shaders/texture_default_projection_vs.glsl",
            fragment_shader=":system:shaders/postprocessing/gaussian_combine_fs.glsl",
        )
        self._quad_fs = geometry.quad_2d_fs()

    def render(self, source, target):
        """Render"""
        # TODO: support source as fbo or texture
        self._cb_luma_program["contrast"] = 4.0
        # self._cb_luma_program['brightness'] = 0.0
        self._cb_luma_buffer.use()
        self._cb_luma_buffer.clear()
        source.use(0)
        self._quad_fs.render(self._cb_luma_program)

        blurred = self._gaussian_1.render(self._cb_luma_buffer.color_attachments[0])
        blurred = self._gaussian_2.render(blurred)
        blurred = self._gaussian_3.render(blurred)

        # Draw combined result to screen
        target.use()
        source.use(0)
        blurred.use(1)
        self._combine_program["color_buffer"] = 0
        self._combine_program["blur_buffer"] = 1
        self._quad_fs.render(self._combine_program)



"""
Simple normal mapping example.

We load a diffuse and normal map and render them using a simple shader.
The normal texture stores a direction vector in the RGB channels
pointing up from the surface.

For simplicity we use the texture coordinates to determine the
screen position but this can be done in other ways.

Controls:
    Mouse: Move light source
    Mouse wheel: Move light source in and out

Run this example from the command line with:
python -m arcade.examples.gl.normal_mapping_simple
"""

import arcade
from arcade.gl import geometry


class NormalMapping(arcade.Window):
    def __init__(self):
        super().__init__(512, 512, "Normal Mapping")

        # Load the color (diffuse) and normal texture
        # These should ideally be the same size
        self.texture_diffuse = self.ctx.load_texture(
            ":resources:images/test_textures/normal_mapping/diffuse.jpg"
        )
        self.texture_normal = self.ctx.load_texture(
            ":resources:images/test_textures/normal_mapping/normal.jpg"
        )

        # Shader program doing basic normal mapping
        self.program = self.ctx.program(
            vertex_shader="""
                #version 330

                // Inputs from the quad_fs geometry
                in vec2 in_vert;
                in vec2 in_uv;

                // Output to the fragment shader
                out vec2 uv;

                void main() {
                    uv = in_uv;
                    gl_Position = vec4(in_vert, 0.0, 1.0);
                }

            """,
            fragment_shader="""
                #version 330

                // Samplers for reading from textures
                uniform sampler2D texture_diffuse;
                uniform sampler2D texture_normal;
                // Global light position we can set from python
                uniform vec3 light_pos;

                // Input from vertex shader
                in vec2 uv;

                // Output to the framebuffer
                out vec4 f_color;

                void main() {
                    // Read RGBA color from the diffuse texture
                    vec4 diffuse = texture(texture_diffuse, uv);
                    // Read normal from RGB channels and convert to a direction vector.
                    // These vectors are like a needle per pixel pointing up from the surface.
                    // Since RGB is 0-1 we need to convert to -1 to 1.
                    vec3 normal = normalize(texture(texture_normal, uv).rgb * 2.0 - 1.0);

                    // Calculate the light direction.
                    // This is the direction between the light position and the pixel position.
                    vec3 light_dir = normalize(light_pos - vec3(uv, 0.0));

                    // Calculate the diffuse factor.
                    // This is the dot product between the light direction and the normal.
                    // It's basically calculating the angle between the two vectors.
                    // The result is a value between 0 and 1.
                    float diffuse_factor = max(dot(normal, light_dir), 0.0);

                    // Write the final color to the framebuffer.
                    // We multiply the diffuse color with the diffuse factor.
                    f_color = vec4(diffuse.rgb * diffuse_factor, 1.0);
                }
            """,
        )
        # Configure what texture channel the samplers should read from
        self.program["texture_diffuse"] = 0
        self.program["texture_normal"] = 1

        # Shortcut for a full screen quad
        # It has two buffers with positions and texture coordinates
        # named "in_vert" and "in_uv" so we need to use that in the vertex shader
        self.quad_fs = geometry.quad_2d_fs()

        # Keep track of mouse coordinates for light position
        self.mouse_x = 0.0
        self.mouse_y = 0.0
        self.mouse_z = 0.25

        self.text = arcade.Text("0, 0, 0", 20, 20, arcade.color.WHITE)

    def on_draw(self):
        self.clear()

        # Bind the textures to the channels we configured in the shader
        self.texture_diffuse.use(0)
        self.texture_normal.use(1)

        # Update the light position uniform variable
        self.program["light_pos"] = self.mouse_x, self.mouse_y, self.mouse_z

        # Run the normal mapping shader (fills a full screen quad)
        self.quad_fs.render(self.program)

        # Draw the mouse coordinates
        self.text.text = f"{self.mouse_x:.2f}, {self.mouse_y:.2f}, {self.mouse_z:.2f}"
        self.text.draw()

    def on_mouse_motion(self, x: int, y: int, dx: int, dy: int):
        """Move the light source with the mouse."""
        # Convert to normalized coordinates
        # (0.0, 0.0) is bottom left, (1.0, 1.0) is top right
        self.mouse_x, self.mouse_y = x / self.width, y / self.height

    def on_mouse_scroll(self, x: int, y: int, scroll_x: float, scroll_y: float):
        """Zoom in/out with the mouse wheel."""
        self.mouse_z += scroll_y * 0.05


NormalMapping().run()



"""
This demo shows using buffered rectangles to draw a grid of squares on the
screen.

For me this runs about 0.002 seconds per frame.

It is faster than demo 1 because we aren't loading the vertices and color
to the card again and again. It could be faster though, if we group all
rectangles together.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.shape_list_demo_2
"""

import arcade
import timeit

WINDOW_WIDTH = 1200
WINDOW_HEIGHT = 800
WINDOW_TITLE = "Shape List Demo 2"

SQUARE_WIDTH = 5
SQUARE_HEIGHT = 5
SQUARE_SPACING = 10


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        super().__init__()

        self.background_color = arcade.color.DARK_SLATE_GRAY

        self.draw_time = 0
        self.shape_list = None

    def setup(self):
        # --- Create the vertex buffers objects for each square before we do
        # any drawing.
        self.shape_list = arcade.shape_list.ShapeElementList()
        for x in range(0, WINDOW_WIDTH, SQUARE_SPACING):
            for y in range(0, WINDOW_HEIGHT, SQUARE_SPACING):
                shape = arcade.shape_list.create_rectangle_filled(
                    center_x=x,
                    center_y=y,
                    width=SQUARE_WIDTH,
                    height=SQUARE_HEIGHT,
                    color=arcade.color.DARK_BLUE,
                )
                self.shape_list.append(shape)

    def on_draw(self):
        """
        Render the screen.
        """

        # This command has to happen before we start drawing
        self.clear()

        # Start timing how long this takes
        draw_start_time = timeit.default_timer()

        # --- Draw all the rectangles
        self.shape_list.draw()

        output = f"Drawing time: {self.draw_time:.3f} seconds per frame."
        arcade.draw_text(output, 20, WINDOW_HEIGHT - 40, arcade.color.WHITE, 18)

        self.draw_time = timeit.default_timer() - draw_start_time


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Reference counters for tracking textures and images in an atlas

* ImageDataRefCounter Tracks how many times an image is used by a texture
  so we can determine when it's safe to remove it.
* UniqueTextureRefCounter Tracks how many times a unique texture is used
  so we can determine when it's safe to remove it. A unique texture is
  simply a texture using the same image and the same vertex order.
"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from arcade.texture import ImageData, Texture


class ImageDataRefCounter:
    """
    Helper class to keep track of how many times an image is used
    by a texture in the atlas to determine when it's safe to remove it.

    Multiple Texture instances can and will use the same ImageData instance.
    """

    def __init__(self) -> None:
        self._data: dict[str, int] = {}
        self._num_decref = 0

    def inc_ref(self, image_data: "ImageData") -> None:
        """
        Increment the reference counter for an image.

        Args:
            image_data: The image to increment the reference counter for
        """
        self._data[image_data.hash] = self._data.get(image_data.hash, 0) + 1

    def dec_ref(self, image_data: "ImageData") -> int:
        """
        Decrement the reference counter for an image returning the new value.

        Args:
            image_data: The image to decrement the reference counter for
        """
        return self.dec_ref_by_hash(image_data.hash)

    def dec_ref_by_hash(self, hash: str) -> int:
        """
        Decrement the reference count for an image by hash returning the new value.

        Raises ``RuntimeError`` if the hash is no longer tracked meaning it doesn't exist
        and/or the reference count is already zero. Otherwise the updated ref counter
        is returned. When 0 is returned we removed the last reference to the image.

        Args:
            hash: The hash of the image to decrement the reference counter for
        """
        val = self._data.get(hash, 0) - 1
        if val < 0:
            raise RuntimeError("Reference counter is already zero or the hash doesn't exist.")

        if val == 0:
            del self._data[hash]
        else:
            self._data[hash] = val

        self._num_decref += 1
        return val

    def get_ref_count(self, image_data: "ImageData") -> int:
        """
        Get the reference count for an image.

        Args:
            image_data: The image to get the reference count for
        """
        return self._data.get(image_data.hash, 0)

    def count_all_refs(self) -> int:
        """Count the total number of references."""
        return sum(self._data.values())

    def get_total_decref(self, reset=True) -> int:
        """
        Get the total number of decrefs.

        This is useful for quickly checking if images was removed from the atlas.
        We can then determine if the atlas needs to be rebuilt. Also for debugging
        purposes.

        Args:
            reset: Reset the decref counter after getting the value
        """
        num_decref = self._num_decref
        if reset:
            self._num_decref = 0
        return num_decref

    def clear(self) -> None:
        """Clear the reference counters"""
        self._data.clear()
        self._num_decref = 0

    def debug_print(self) -> None:
        """print the keys and counters for debugging purposes."""
        print(f"{self.__class__.__name__}:")
        for key, val in self._data.items():
            print(f"  {key}: {val}")

    def __len__(self) -> int:
        return len(self._data)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} ref_count={self.count_all_refs()} data={self._data}>"


class UniqueTextureRefCounter:
    """
    Helper class to keep track of how many times a unique texture is used.

    A "unique texture" is based on the ``atlas_name`` of the texture meaning
    a texture using the same image and the same vertex order.
    """

    def __init__(self) -> None:
        self._data: dict[str, int] = {}
        self._num_decref = 0

    def inc_ref(self, image_data: "Texture") -> None:
        """
        Increment the reference count for an image.

        Args:
            image_data: The image to increment the reference counter for
        """
        self._data[image_data.atlas_name] = self._data.get(image_data.atlas_name, 0) + 1

    def dec_ref(self, image_data: "Texture") -> int:
        """
        Decrement the reference counter for an image returning the new value.

        Args:
            image_data: The image to decrement the reference counter for
        """
        return self.dec_ref_by_atlas_name(image_data.atlas_name)

    def dec_ref_by_atlas_name(self, atlas_name: str) -> int:
        """
        Decrement the reference counter for an image by name returning the new value.

        Raises ``RuntimeError`` if the atlas name is no longer tracked meaning it doesn't exist
        and/or the reference count is already zero. Otherwise the updated ref counter
        is returned. When 0 is returned we removed the last reference to the texture.

        Args:
            atlas_name: The name of the texture to decrement the reference counter
        """
        val = self._data.get(atlas_name, 0) - 1
        if val < 0:
            raise RuntimeError("Reference counter is already zero or the atlas name doesn't exist.")

        if val == 0:
            del self._data[atlas_name]
        else:
            self._data[atlas_name] = val

        self._num_decref += 1
        return val

    def get_ref_count(self, image_data: "ImageData") -> int:
        """
        Get the reference counter for an image.

        Args:
            image_data: The image to get the reference count for
        """
        return self._data.get(image_data.hash, 0)

    def count_all_refs(self) -> int:
        """Helper function to count the total number of references."""
        return sum(self._data.values())

    def get_total_decref(self, reset=True) -> int:
        """
        Get the total number of decrefs.

        This is useful to keep track of how many textures was removed from the atlas
        either for functional or debugging purposes.

        Args:
            reset: Reset the counter after getting the value
        """
        num_decref = self._num_decref
        if reset:
            self._num_decref = 0
        return num_decref

    def clear(self) -> None:
        """Clear the reference counters"""
        self._data.clear()
        self._num_decref = 0

    def debug_print(self) -> None:
        """Print all the keys and counters for debugging purposes"""
        print(f"{self.__class__.__name__}:")
        for key, val in self._data.items():
            print(f"  {key}: {val}")

    def __len__(self) -> int:
        return len(self._data)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} ref_count={self.count_all_refs()} data={self._data}>"



from .sprite_list import SpriteList, SpriteSequence
from .spatial_hash import SpatialHash
from .collision import (
    get_distance_between_sprites,
    get_closest_sprite,
    check_for_collision,
    check_for_collision_with_list,
    check_for_collision_with_lists,
    get_sprites_at_point,
    get_sprites_at_exact_point,
    get_sprites_in_rect,
)


__all__ = [
    "SpriteList",
    "SpriteSequence",
    "SpatialHash",
    "get_distance_between_sprites",
    "get_closest_sprite",
    "check_for_collision",
    "check_for_collision_with_list",
    "check_for_collision_with_lists",
    "get_sprites_at_point",
    "get_sprites_at_exact_point",
    "get_sprites_in_rect",
]



"""
Simple 3D Example

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.3d_cube
"""

from pyglet.math import Mat4, Vec3
import arcade
from arcade.gl import geometry


class GameView(arcade.Window):

    def __init__(self, width, height, title):
        super().__init__(width, height, title, resizable=True)
        self.set_vsync(True)
        # Use the standard cube
        self.cube = geometry.cube()
        # Simple color lighting program
        self.program = self.ctx.program(
            vertex_shader="""
            #version 330

            uniform mat4 projection;
            uniform mat4 modelview;

            in vec3 in_position;
            in vec3 in_normal;

            out vec3 normal;
            out vec3 pos;

            void main() {
                vec4 p = modelview * vec4(in_position, 1.0);
                gl_Position = projection * p;
                mat3 m_normal = transpose(inverse(mat3(modelview)));
                normal = m_normal * in_normal;
                pos = p.xyz;
            }
            """,
            fragment_shader="""
            #version 330

            out vec4 fragColor;

            in vec3 normal;
            in vec3 pos;

            void main()
            {
                float l = dot(normalize(-pos), normalize(normal));
                fragColor = vec4(1.0) * (0.25 + abs(l) * 0.75);
            }
            """,
        )
        self.on_resize(*self.get_size())

    def on_draw(self):
        self.clear()
        self.ctx.enable_only(self.ctx.CULL_FACE, self.ctx.DEPTH_TEST)

        translate = Mat4.from_translation(Vec3(0, 0, -1.75))
        rx = Mat4.from_rotation(self.time, Vec3(1, 0, 0))
        ry = Mat4.from_rotation(self.time * 0.77, Vec3(0, 1, 0))
        modelview = translate @ rx @ ry
        self.program["modelview"] = modelview

        self.cube.render(self.program)

    def on_resize(self, width, height):
        """Set up viewport and projection"""
        self.ctx.viewport = 0, 0, width, height
        self.program["projection"] = (
            Mat4.perspective_projection(self.aspect_ratio, 0.1, 100, fov=60)
        )


if __name__ == "__main__":
    GameView(1280, 720, "3D Cube")
    arcade.run()



"""
This demo shows the speed of drawing a full grid of squares using no buffering.

For me this takes about 0.16 seconds per frame.

It is slow because we load all the points and all the colors to the card every
time.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.shape_list_demo_1
"""

import arcade
import timeit

WINDOW_WIDTH = 1200
WINDOW_HEIGHT = 800
WINDOW_TITLE = "Shape List Demo 1"

SQUARE_WIDTH = 5
SQUARE_HEIGHT = 5
SQUARE_SPACING = 10


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        super().__init__()

        self.background_color = arcade.color.DARK_SLATE_GRAY

        self.draw_time = 0

    def on_draw(self):
        """
        Render the screen.
        """

        # This command has to happen before we start drawing
        self.clear()

        # Start timing how long this takes
        draw_start_time = timeit.default_timer()

        # --- Draw all the rectangles
        for x in range(0, WINDOW_WIDTH, SQUARE_SPACING):
            for y in range(0, WINDOW_HEIGHT, SQUARE_SPACING):
                arcade.draw_rect_filled(arcade.rect.XYWH(x, y, SQUARE_WIDTH, SQUARE_HEIGHT),
                                        arcade.color.DARK_BLUE)

        # Print the timing
        output = f"Drawing time: {self.draw_time:.3f} seconds per frame."
        arcade.draw_text(output, 20, WINDOW_HEIGHT - 40, arcade.color.WHITE, 18)

        self.draw_time = timeit.default_timer() - draw_start_time


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from typing import TypeVar

from arcade import View
from arcade.gui.ui_manager import UIManager
from arcade.gui.widgets import UIWidget

W = TypeVar("W", bound=UIWidget)


class UIView(View):
    """This view provides basic GUI setup.

    This is a convenience class, which adds the UIManager to the view under `self.ui`.
    The UIManager is enabled when the view is shown and disabled when the view is hidden.

    This class provides two draw callbacks: on_draw_before_ui and on_draw_after_ui.
    Use these to draw custom elements before or after the UI elements are drawn.

    The screen is cleared before on_draw_before_ui is called
    with the background color of the window.

    If you override ``on_show_view`` or ``on_hide_view``,
    don't forget to call super().on_show_view() or super().on_hide_view().

    """

    def __init__(self):
        super().__init__()
        self.ui = UIManager()
        """
        The UIManager of this view.
        """

    def add_widget(self, widget: W, *, index=None, layer=UIManager.DEFAULT_LAYER) -> W:
        """Add a widget to the UIManager of this view.

        Args:
            widget: widget to add
            index: position a widget is added, None has the highest priority
            layer: layer which the widget should be added to, higher layer are above
        """
        return self.ui.add(widget, index=index, layer=layer)

    def on_show_view(self):
        """If subclassing UIView, don't forget to call super().on_show_view()."""
        self.ui.enable()

    def on_hide_view(self):
        """If subclassing UIView, don't forget to call super().on_hide_view()."""
        self.ui.disable()

    def on_draw(self):
        """To subclass UIView and add custom drawing, override on_draw_before_ui
        and on_draw_after_ui.
        """
        self.clear()
        self.on_draw_before_ui()
        self.ui.draw()
        self.on_draw_after_ui()

    def on_draw_before_ui(self):
        """Use this method to draw custom elements before the UI elements are drawn."""
        pass

    def on_draw_after_ui(self):
        """Use this method to draw custom elements after the UI elements are drawn."""
        pass



"""Prevents import issues.

If an :py:mod:`arcade.types` submodule attempts to run
``from arcade.types import AsFloat``, it could cause issues with
circular imports or partially initialized modules.
"""

#: 1. Makes pyright happier while also telling readers
#: 2. Tells readers we're converting any ints to floats
AsFloat = float | int

__all__ = ["AsFloat"]



import arcade
from arcade.experimental.shadertoy import Shadertoy

# Do the math to figure out our screen dimensions
SCREEN_WIDTH = 1280
SCREEN_HEIGHT = 720
SCREEN_TITLE = "ShaderToy Demo"


class MyGame(arcade.Window):
    def __init__(self, width, height, title):
        super().__init__(width, height, title, resizable=True)
        self.shadertoy = Shadertoy(
            self.get_framebuffer_size(),
            """
                void mainImage( out vec4 fragColor, in vec2 fragCoord )
                {
                    // Calculate the texture coordinate of the current fragment.
                    // This interpolates from 0,0 to 1,1 from lower left to upper right
                    vec2 uv = fragCoord.xy / iResolution.xy;
                    // Write the fragment color
                    fragColor = vec4(uv, 0.0, 1.0);
                }
            """,
        )

    def on_draw(self):
        self.clear()
        self.shadertoy.render()


if __name__ == "__main__":
    MyGame(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_TITLE)
    arcade.run()



"""
Load level data in the background with interactive previews.

Level preview borders will turn green when their data loads:

1. Pan the camera by clicking and dragging
2. Zoom in and out by scrolling up or down

Loading data during gameplay always risks slowdowns. These risks grow
grow with number, size, and loading complexity of files. Some games
avoid the problem by using non-interactive loading screens. This example
uses a different approach.

Background loading works if a game has enough RAM and CPU cycles to
run light features while loading. These can be the UI or menus, or even
gameplay light enough to avoid interfering with the loading thread. For
example, players can handle inventory or communication while data loads.

Although Python's threading module has many pitfalls, we'll avoid them
by keeping things simple:

1. There is only one background loader thread
2. The loader thread will load each map in order, one after another
3. If a map loads successfully, the UI an interactive preview

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.threaded_loading
"""
import sys
from time import sleep

# Python's threading module has proven tools for working with threads, and
# veteran developers may want to explore 3.13's new 'No-GIL' concurrency.
import threading

import arcade
from arcade.color import RED, GREEN, BLUE, WHITE
from arcade.math import clamp

# Window size and title
WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = 'Threaded Tilemap Loading'

# We'll simulate loading large files by adding a loading delay to
# each map we load. You can omit the delay in your own projects.
ARTIFICIAL_DELAY = 1


# The resource handle prefix to load map files from
LEVEL_LOCATION = ':assets:tiled_maps/'
# Level filenames in the resource folder
LEVELS = (
    'test_map_1.json',
    'test_map_2.json',
    'test_map_3.json',
    'test_map_4.json',  # Doesn't exist so we can simulate failed reads
    'test_map_5.json',  # Intentionally empty file
    'test_map_6.json',
    'test_map_7.json',
)

# Rendering layout controls
COLUMN_COUNT = 4
LEVEL_RENDERER_SIZE = WINDOW_WIDTH // 5 - 10, WINDOW_HEIGHT // 5 - 10


class LevelLoader:
    """Wrap a loader thread which runs level loading in the background.

    IMPORTANT: NEVER call graphics code from threads! They break OpenGL!

    It's common to group threading tasks into manager objects which track
    and coordinate them. Advanced thread managers often keep idle threads
    'alive' to re-use when needed. These complex techniques are beyond the
    scope of this tutorial.
    """

    def __init__(self, levels: tuple[str, ...], location: str):
        self._levels = levels
        self._location = location

        self._begun: bool = False
        self._finished: bool = False

        # Threads do not start until their `start` method is called.
        self.loading_thread = threading.Thread(target=self._load_levels)

        self._loaded_levels: dict[str, arcade.TileMap] = {}
        self._failed_levels: set[str] = set()
        self._current_level: str = ''

        # Avoid the difficulties of coordinating threads without
        # freezing by using one loading thread with a one lock.
        self._interaction_lock = threading.Lock()

    # An underscore at the start of a name is how a developer tells
    # others to treat things as private. Here, it means that only
    # LevelLoader should ever call `_load_levels` directly.
    def _load_levels(self):
        for level in self._levels:
            with self._interaction_lock:
                self._current_level = level

            sleep(ARTIFICIAL_DELAY)  # "Slow" down (delete this line before use)

            # Since unhandled exceptions "kill" threads, we catch the only major
            # exception we expect. Level 4 is intentionally missing to test cases
            # such as this one when building map loading code.
            try:
                path = f'{self._location}{level}'
                tilemap = arcade.load_tilemap(path, lazy=True)
            except FileNotFoundError:
                print(f"ERROR: {level} doesn't exist, skipping!", file=sys.stderr)
                with self._interaction_lock:
                    self._failed_levels.add(level)
                continue

            with self._interaction_lock:
                self._loaded_levels[level] = tilemap

        with self._interaction_lock:
            self._finished = True

    def start_loading_levels(self):
        with self._interaction_lock:
            if not self._begun:
                self.loading_thread.start()
                self._begun = True

    @property
    def current_level(self) -> str:
        with self._interaction_lock:
            return self._current_level

    @property
    def begun(self):
        with self._interaction_lock:
            return self._begun

    @property
    def finished(self):
        with self._interaction_lock:
            return self._finished

    def is_level_loaded(self, level: str) -> bool:
        with self._interaction_lock:
            return level in self._loaded_levels

    def did_level_fail(self, level: str) -> bool:
        with self._interaction_lock:
            return level in self._failed_levels

    def get_level(self, level: str) -> arcade.TileMap | None:
        with self._interaction_lock:
            return self._loaded_levels.get(level, None)


class LevelRenderer:
    """
    Draws previews of loaded data and colored borders to show status.
    """

    def __init__(
            self,
            level: str,
            level_loader: LevelLoader,
            location: arcade.types.Point2,
            size: tuple[int, int]
        ):
        self.level_name = level
        self.loader = level_loader

        self.location = location
        self.size = size
        x, y = location
        self.camera: arcade.Camera2D = arcade.Camera2D(
            arcade.XYWH(x, y, size[0], size[1])
        )
        camera_x, camera_y = self.camera.position
        self.level: arcade.TileMap | None = None
        self.level_text: arcade.Text = arcade.Text(
            level,
            camera_x, camera_y,
            anchor_x='center',
            anchor_y='center'
        )

    def update(self):
        level = self.level
        loader = self.loader
        if not level and loader.is_level_loaded(self.level_name):
            self.level = self.loader.get_level(self.level_name)

    def draw(self):
        # Activate the camera to render into its viewport rectangle
        with self.camera.activate():
            if self.level:
                for spritelist in self.level.sprite_lists.values():
                    spritelist.draw()
            self.level_text.draw()

        # Choose a color based on the load status
        if self.level is not None:
            color = GREEN
        elif self.loader.did_level_fail(self.level_name):
            color = RED
        elif self.loader.current_level == self.level_name:
            color = BLUE
        else:
            color = WHITE

        # Draw the outline over any thumbnail
        arcade.draw_rect_outline(self.camera.viewport, color, 3)

    def point_in_area(self, x, y):
        return self.camera.viewport.point_in_rect((x, y))

    def drag(self, dx, dy):
        # Store a few values locally to make the math easier to read
        camera = self.camera
        x, y = camera.position
        zoom = camera.zoom

        # Move the camera while accounting for zoom
        camera.position = x - dx / zoom, y - dy / zoom

    def scroll(self, scroll):
        camera = self.camera
        zoom = camera.zoom
        camera.zoom = clamp(zoom + scroll / 10, 0.1, 10)


class GameView(arcade.View):

    def __init__(self, window = None, background_color = None):
        super().__init__(window, background_color)
        self.level_loader = LevelLoader(LEVELS, LEVEL_LOCATION)
        self.level_renderers: list[LevelRenderer] = []

        for idx, level in enumerate(LEVELS):
            row = idx // COLUMN_COUNT
            column = idx % COLUMN_COUNT
            pos = (1 + column) / 5 * self.width, (3 - row) / 4 * self.height
            self.level_renderers.append(
                LevelRenderer(level, self.level_loader, pos, LEVEL_RENDERER_SIZE)
            )

        self.loading_sprite = arcade.SpriteSolidColor(
            64,
            64,
            self.center_x,
            200,
            WHITE
        )

        self.dragging = None

    def on_show_view(self):
        self.level_loader.start_loading_levels()

    def on_update(self, delta_time):
        # This sprite will spin one revolution per second. Even when loading levels this
        # won't freeze thanks to the threaded loading
        self.loading_sprite.angle = (360 * self.window.time) % 360
        for renderer in self.level_renderers:
            renderer.update()

        if self.dragging is not None:
            self.window.set_mouse_cursor(
                self.window.get_system_mouse_cursor(self.window.CURSOR_SIZE))
        else:
            self.window.set_mouse_cursor(None)

    def on_draw(self):
        self.clear()
        arcade.draw_sprite(self.loading_sprite)
        for renderer in self.level_renderers:
            renderer.draw()

    def on_mouse_press(self, x, y, button, modifiers):
        for renderer in self.level_renderers:
            if renderer.point_in_area(x, y):
                self.dragging = renderer
                break

    def on_mouse_release(self, x, y, button, modifiers):
        self.dragging = None

    def on_mouse_drag(self, x, y, dx, dy, _buttons, _modifiers):
        if self.dragging is not None:
            self.dragging.drag(dx, dy)

    def on_mouse_scroll(self, x, y, scroll_x, scroll_y):
        if self.dragging is not None:
            self.dragging.scroll(scroll_y)
            return
        for renderer in self.level_renderers:
            if renderer.point_in_area(x, y):
                renderer.scroll(scroll_y)
                break


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""Contains experimental GUI elements.

The API of these components may change within minor version updates.
No Deprecation warnings are given for changes in this module.
"""

from arcade.gui.experimental.scroll_area import UIScrollArea
from arcade.gui.experimental.password_input import UIPasswordInput

__all__ = [
    "UIScrollArea",
    "UIPasswordInput",
]



"""
Move Sprite With Keyboard

Simple program to show moving a sprite with the keyboard.
The sprite_move_keyboard_better.py example is slightly better
in how it works, but also slightly more complex.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_move_keyboard
"""

import arcade

SPRITE_SCALING = 0.5

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Move Sprite with Keyboard Example"

MOVEMENT_SPEED = 5


class Player(arcade.Sprite):
    """ Player Class """

    def update(self, delta_time: float = 1/60):
        """ Move the player """
        # Move player.
        # Remove these lines if physics engine is moving player.
        self.center_x += self.change_x
        self.center_y += self.change_y

        # Check for out-of-bounds
        if self.left < 0:
            self.left = 0
        elif self.right > WINDOW_WIDTH - 1:
            self.right = WINDOW_WIDTH - 1

        if self.bottom < 0:
            self.bottom = 0
        elif self.top > WINDOW_HEIGHT - 1:
            self.top = WINDOW_HEIGHT - 1


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        """
        Initializer
        """

        # Call the parent class initializer
        super().__init__()

        # Variables that will hold sprite lists
        self.player_list = None

        # Set up the player info
        self.player_sprite = None

        # Set the background color
        self.background_color = arcade.color.AMAZON

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Sprite lists
        self.player_list = arcade.SpriteList()

        # Set up the player
        self.player_sprite = Player(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=SPRITE_SCALING,
        )
        self.player_sprite.center_x = 50
        self.player_sprite.center_y = 50
        self.player_list.append(self.player_sprite)

    def on_draw(self):
        """
        Render the screen.
        """

        # This command has to happen before we start drawing
        self.clear()

        # Draw all the sprites.
        self.player_list.draw()

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Move the player
        self.player_list.update(delta_time)

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed. """

        # If the player presses a key, update the speed
        if key == arcade.key.UP:
            self.player_sprite.change_y = MOVEMENT_SPEED
        elif key == arcade.key.DOWN:
            self.player_sprite.change_y = -MOVEMENT_SPEED
        elif key == arcade.key.LEFT:
            self.player_sprite.change_x = -MOVEMENT_SPEED
        elif key == arcade.key.RIGHT:
            self.player_sprite.change_x = MOVEMENT_SPEED

    def on_key_release(self, key, modifiers):
        """Called when the user releases a key. """

        # If a player releases a key, zero out the speed.
        # This doesn't work well if multiple keys are pressed.
        # Use 'better move by keyboard' example if you need to
        # handle this.
        if key == arcade.key.UP or key == arcade.key.DOWN:
            self.player_sprite.change_y = 0
        elif key == arcade.key.LEFT or key == arcade.key.RIGHT:
            self.player_sprite.change_x = 0


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
A helper class to keep track of texture coordinates stored in a texture.
"""

from __future__ import annotations

from array import array
from collections import deque
from typing import TYPE_CHECKING

from .base import (
    UV_TEXTURE_WIDTH,
    TexCoords,
)

if TYPE_CHECKING:
    from arcade import ArcadeContext
    from arcade.gl import Texture2D


class UVData:
    """
    A container for float32 texture coordinates stored in a texture.
    Each texture coordinate has a slot/index in the texture and is
    looked up by a shader to obtain the texture coordinates. The
    shader would look up texture coordinates by the id of the texture.

    The purpose of this system is to:

    - Greatly increase the performance of the texture atlas
    - Greatly simplify the system
    - Allow images to move freely around the atlas without having to update the vertex buffers.
      Meaning we can allow re-building and re-sizing. The resize can even
      be done in the GPU by rendering the old atlas into the new one.
    - Avoid spending lots of time packing texture data into buffers
    - Avoid spending lots of buffer memory

    Args:
        ctx:
            The Arcade context
        capacity:
            The number of textures the atlas keeps track of.
            This is multiplied by 4096. Meaning capacity=2 is 8192 textures.
    """

    def __init__(self, ctx: ArcadeContext, capacity: int):
        self._ctx = ctx
        self._capacity = capacity
        self._num_slots = UV_TEXTURE_WIDTH * capacity
        self._dirty = False

        # The GPU resource
        self._texture = self._ctx.texture(
            (UV_TEXTURE_WIDTH, self._num_slots * 2 // UV_TEXTURE_WIDTH),
            components=4,
            dtype="f4",
        )
        self._texture.filter = self._ctx.NEAREST, self._ctx.NEAREST

        # Python resources: data + tracker for slots
        # 8 floats per texture (4 x vec2 coordinates)
        self._data = array("f", [0] * self._num_slots * 8)
        self._slots: dict[str, int] = dict()
        self._slots_free = deque(i for i in range(0, self._num_slots))

    def clone_with_slots(self) -> UVData:
        """
        Clone the UVData with the current slot data. The UV data itself
        and the opengl texture is not cloned.

        This is useful when we are re-building the atlas since we can't
        lost the slot data since the indices in the uv texture must not
        change. If they change the entire spritelist must be updated.
        """
        clone = UVData(self._ctx, self._capacity)
        clone._slots = self._slots
        clone._slots_free = self._slots_free
        clone._dirty = True
        return clone

    @property
    def num_slots(self) -> int:
        """The amount of texture coordinates (x4) this UVData can hold"""
        return self._num_slots

    @property
    def num_free_slots(self) -> int:
        """The amount of free texture coordinates slots"""
        return len(self._slots_free)

    @property
    def texture(self) -> Texture2D:
        """The opengl texture containing the texture coordinates"""
        return self._texture

    def get_slot_or_raise(self, name: str) -> int:
        """
        Get the slot for a texture by name or raise an exception

        Args:
            name: The name of the texture
        """
        slot = self._slots.get(name)
        if slot is None:
            raise Exception(f"Texture '{name}' not found in UVData")
        return slot

    def get_existing_or_free_slot(self, name: str) -> int:
        """
        Get the slot for a texture by name or a free slot.
        Getting existing slots is useful when the resize or re-build
        the atlas.

        Args:
            name: The name of the texture
        """
        slot = self._slots.get(name)
        if slot is not None:
            return slot

        try:
            slot = self._slots_free.popleft()
            self._slots[name] = slot
            return slot
        except IndexError:
            raise Exception(
                "No more free slots in the UV texture."
                f"Max number of textures: {self._num_slots}."
                "Consider creating a texture atlas with a larger capacity."
            )

    def free_slot_by_name(self, name: str) -> None:
        """
        Free a slot for a texture by name.
        If the slot is not found no action is taken

        Args:
            name: The name of the texture
        """
        try:
            slot = self._slots.pop(name)
            self._slots_free.appendleft(slot)
        except KeyError:
            return

    def set_slot_data(self, slot: int, data: TexCoords) -> None:
        """
        Update the texture coordinates for a slot.

        Args:
            slot: The slot to update
            data: The texture coordinates
        """
        self._data[slot * 8 : slot * 8 + 8] = array("f", data)
        self._dirty = True

    def write_to_texture(self) -> None:
        """Write the texture coordinates to the opengl texture if dirty"""
        if self._dirty:
            self._texture.write(self._data, 0)
            self._dirty = False

    def __len__(self) -> int:
        return len(self._slots)



"""
Utility functions to keep performance information
"""

import time
from collections import deque

import pyglet

# Evil globals
_timings: dict[str, deque[float]] = {}
_pyglets_dispatch_event = None
_frame_times: deque = deque()
_max_history: int = 100

__all__ = [
    "print_timings",
    "clear_timings",
    "get_timings",
    "enable_timings",
    "disable_timings",
    "get_fps",
    "timings_enabled",
]


def _dispatch_event(self, *args) -> None:
    """
    This function will be monkey-patched over Pyglet's dispatch event function.
    """
    # Name of the dispatched event, like 'on_draw'
    name = args[0]

    # Start the clock, and keep the time if this is on_draw for FPS calcs.
    start_time = time.perf_counter()
    if name == "on_draw":
        _frame_times.append(start_time)

    # Call Pyglet's dispatch event function
    if _pyglets_dispatch_event is not None:
        _pyglets_dispatch_event(self, *args)

    # Stop the clock
    end_time = time.perf_counter()
    processing_time = end_time - start_time

    # Get the historical list of timings, or create one if we don't have it.
    if name in _timings:
        data = _timings[name]
    else:
        data = deque()
        _timings[name] = data

    # Add out time to the list
    data.append(processing_time)

    # Past out history limit? Pop off the first one on the list
    if len(data) > _max_history:
        data.popleft()


def print_timings() -> None:
    """
    Print event handler statistics to stdout as a table.

    Performance tracking must be enabled with
    :func:`arcade.enable_timings` before calling this function.

    See :ref:`performance_statistics_example` for an example of how to
    use function.

    The statistics consist of:

    * how many times each registered event was called
    * the average time for handling each type of event in seconds

    The table looks something like:

    .. code-block:: text

        Event          Count Average Time
        -------------- ----- ------------
        on_update         60       0.0000
        on_mouse_enter     1       0.0000
        on_mouse_motion   39       0.0000
        on_expose          1       0.0000
        on_draw           60       0.0020
    """
    global _timings
    print()
    print("Event          Count Average Time")
    print("-------------- ----- ------------")
    for index in _timings:
        data = _timings[index]

        call_count = len(data)
        average_time = sum(data) / len(data)
        print(f"{index:15}{call_count:5}  {average_time:11.4f}")


def clear_timings() -> None:
    """
    Reset the count & average time for each event type to zero.

    Performance tracking must be enabled with
    :func:`arcade.enable_timings` before calling this function.

    See :ref:`performance_statistics_example` for an example of how to
    use function.
    """
    global _timings
    _timings = {}


def get_timings() -> dict[str, deque[float]]:
    """
    Get a dict of the current dispatch event timings.

    Performance tracking must be enabled with
    :func:`arcade.enable_timings` before calling this function.

    :return: A dict of event timing data, consisting of counts and
             average handler duration.
    """
    return _timings


def enable_timings(max_history: int = 100) -> None:
    """
    Enable recording of performance information.

    This function must be called before using any other performance
    features, except for :func:`arcade.timings_enabled`, which can
    be called at any time.

    See :ref:`performance_statistics_example` for an example of how to
    use function.

    Args:
        max_history:
            How many frames to keep performance info for.
    """
    global _pyglets_dispatch_event, _max_history

    if pyglet.window.BaseWindow.dispatch_event == _dispatch_event:
        raise ValueError("Timings already enabled.")

    # Save the original pyglet dispatch event function
    _pyglets_dispatch_event = pyglet.window.BaseWindow.dispatch_event

    # Override the pyglet dispatch event function
    pyglet.window.BaseWindow.dispatch_event = _dispatch_event  # type: ignore
    _max_history = max_history


def disable_timings() -> None:
    """
    Disable collection of timing information.

    Performance tracking must be enabled with
    :func:`arcade.enable_timings` before calling this function.
    """
    if pyglet.window.BaseWindow.dispatch_event != _dispatch_event:
        raise ValueError("Timings are not enabled.")

    # Restore the original pyglet dispatch event function
    pyglet.window.BaseWindow.dispatch_event = _pyglets_dispatch_event  # type: ignore

    clear_timings()


def get_fps(frame_count: int = 60) -> float:
    """
    Get the FPS over the last ``frame_count`` frames.

    Performance tracking must be enabled with
    :func:`arcade.enable_timings` before calling this function.

    To get the FPS over the last 30 frames, you would pass 30 instead
    of the default 60.

    See :ref:`performance_statistics_example` for an example of how to
    use function.

    Args:
        frame_count: How many frames to calculate the FPS over.
    """
    cur_time = time.perf_counter()
    if len(_frame_times) == 0:
        return 0
    if len(_frame_times) < frame_count:
        frame_count = len(_frame_times)

    start_time = _frame_times[-frame_count]
    total_time = cur_time - start_time
    fps = frame_count / total_time
    return fps


def timings_enabled() -> bool:
    """
    Return true if timings are enabled, false otherwise.

    This function can be used at any time to check if timings are
    enabled. See :func:`arcade.enable_timings` for more information.

    :return: Whether timings are currently enabled.
    """
    return pyglet.window.BaseWindow.dispatch_event == _dispatch_event



from .video_player import VideoPlayer

__all__ = ["VideoPlayer"]

# Keep cv2 an optional dependency
try:
    from .video_cv2 import VideoPlayerCV2  # noqa: F401
    from .video_record_cv2 import VideoRecorderCV2  # noqa: F401

    __all__.extend(["VideoPlayerCV2", "VideoRecorderCV2"])


except ImportError:
    pass



"""
Asteroid Smasher

Shoot space rocks in this demo program created with Python and the
Arcade library.

Artwork from https://kenney.nl

For a fancier example of this game, see:
https://github.com/pythonarcade/asteroids

If Python and Arcade are installed, this example can be run from
the command line with:
python -m arcade.examples.asteroid_smasher
"""
import random
import math
import arcade

from typing import cast

WINDOW_TITLE = "Asteroid Smasher"
STARTING_ASTEROID_COUNT = 3
SCALE = 0.5

# Screen dimensions and limits
WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
OFFSCREEN_SPACE = 300
LEFT_LIMIT = -OFFSCREEN_SPACE
RIGHT_LIMIT = WINDOW_WIDTH + OFFSCREEN_SPACE
BOTTOM_LIMIT = -OFFSCREEN_SPACE
TOP_LIMIT = WINDOW_HEIGHT + OFFSCREEN_SPACE

# Control player speed
TURN_SPEED = 3
THRUST_AMOUNT = 0.2

# Asteroid types
ASTERIOD_TYPE_BIG = 4
ASTERIOD_TYPE_MEDIUM = 3
ASTERIOD_TYPE_SMALL = 2
ASTERIOD_TYPE_TINY = 1


class TurningSprite(arcade.Sprite):
    """ Sprite that sets its angle to the direction it is traveling in. """
    def update(self, delta_time=1 / 60):
        """ Move the sprite """
        super().update(delta_time)
        self.angle = -math.degrees(math.atan2(self.change_y, self.change_x))


class ShipSprite(arcade.Sprite):
    """ Sprite that represents our spaceship. """
    def __init__(self, filename, scale):
        """ Set up the spaceship. """

        # Call the parent Sprite constructor
        super().__init__(filename, scale=scale)

        # Info on the space ship.
        # Angle comes in automatically from the parent class.
        self.thrust = 0
        self.speed = 0
        self.max_speed = 4
        self.drag = 0.05
        self.respawning = 0

        # Mark that we are respawning.
        self.respawn()

    def respawn(self):
        """
        Called when we die and need to make a new ship.
        'respawning' is an invulnerability timer.
        """
        # If we are in the middle of respawning, this is non-zero.
        self.respawning = 1
        self.alpha = 0
        self.center_x = WINDOW_WIDTH / 2
        self.center_y = WINDOW_HEIGHT / 2
        self.angle = 0

    def update(self, delta_time=1 / 60):
        """ Update our position and other particulars. """

        # Is the user spawning
        if self.respawning:
            # Increase spawn counter, setting alpha to that amount
            self.respawning += 1
            self.alpha = self.respawning
            # Once we are close enough, set alpha to 255 and clear
            # respawning flag
            if self.respawning > 230:
                self.respawning = 0
                self.alpha = 255

        # Apply drag forward
        if self.speed > 0:
            self.speed -= self.drag
            if self.speed < 0:
                self.speed = 0
        # Apply drag reverse
        if self.speed < 0:
            self.speed += self.drag
            if self.speed > 0:
                self.speed = 0

        # Apply thrust
        self.speed += self.thrust

        # Enforce speed limit
        if self.speed > self.max_speed:
            self.speed = self.max_speed
        if self.speed < -self.max_speed:
            self.speed = -self.max_speed

        # Calculate movement vector based on speed/angle
        self.change_x = math.sin(math.radians(self.angle)) * self.speed
        self.change_y = math.cos(math.radians(self.angle)) * self.speed

        # Apply movement vector
        self.center_x += self.change_x
        self.center_y += self.change_y

        # If the ship goes off-screen, move it to the other side of the window
        if self.right < 0:
            self.left = WINDOW_WIDTH
        if self.left > WINDOW_WIDTH:
            self.right = 0
        if self.bottom < 0:
            self.top = WINDOW_HEIGHT
        if self.top > WINDOW_HEIGHT:
            self.bottom = 0

        """ Call the parent class. """
        super().update()


class AsteroidSprite(arcade.Sprite):
    """Sprite that represents an asteroid."""

    def __init__(self, image_file_name, scale, type):
        super().__init__(image_file_name, scale=scale)
        self.type = type

    def update(self, delta_time=1 / 60):
        """ Move the asteroid around. """
        super().update(delta_time)
        if self.center_x < LEFT_LIMIT:
            self.center_x = RIGHT_LIMIT
        if self.center_x > RIGHT_LIMIT:
            self.center_x = LEFT_LIMIT
        if self.center_y > TOP_LIMIT:
            self.center_y = BOTTOM_LIMIT
        if self.center_y < BOTTOM_LIMIT:
            self.center_y = TOP_LIMIT


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        super().__init__()

        self.game_over = False

        # Create sprite lists
        self.player_sprite_list = arcade.SpriteList()
        self.asteroid_list = arcade.SpriteList()
        self.bullet_list = arcade.SpriteList()
        self.ship_life_list = arcade.SpriteList()

        # Set up the player
        self.score = 0
        self.player_sprite = None
        self.lives = 3

        # Load sounds
        self.laser_sound = arcade.load_sound(":resources:sounds/hurt5.wav")
        self.hit_sound1 = arcade.load_sound(":resources:sounds/explosion1.wav")
        self.hit_sound2 = arcade.load_sound(":resources:sounds/explosion2.wav")
        self.hit_sound3 = arcade.load_sound(":resources:sounds/hit1.wav")
        self.hit_sound4 = arcade.load_sound(":resources:sounds/hit2.wav")

        # Text fields
        self.text_score = arcade.Text(
            f"Score: {self.score}",
            x=10,
            y=70,
            font_size=13,
        )
        self.text_asteroid_count = arcade.Text(
            f"Asteroid Count: {len(self.asteroid_list)}",
            x=10,
            y=50,
            font_size=13,
        )

    def start_new_game(self):
        """ Set up the game and initialize the variables. """

        self.game_over = False

        # Sprite lists
        self.player_sprite_list = arcade.SpriteList()
        self.asteroid_list = arcade.SpriteList()
        self.bullet_list = arcade.SpriteList()
        self.ship_life_list = arcade.SpriteList()

        # Set up the player
        self.score = 0
        self.player_sprite = ShipSprite(
            ":resources:images/space_shooter/playerShip1_orange.png",
            scale=SCALE,
        )
        self.player_sprite_list.append(self.player_sprite)
        self.lives = 3

        # Set up the little icons that represent the player lives.
        cur_pos = 10
        for i in range(self.lives):
            life = arcade.Sprite(
                ":resources:images/space_shooter/playerLife1_orange.png",
                scale=SCALE,
            )
            life.center_x = cur_pos + life.width
            life.center_y = life.height
            cur_pos += life.width
            self.ship_life_list.append(life)

        # Make the asteroids
        image_list = (
            ":resources:images/space_shooter/meteorGrey_big1.png",
            ":resources:images/space_shooter/meteorGrey_big2.png",
            ":resources:images/space_shooter/meteorGrey_big3.png",
            ":resources:images/space_shooter/meteorGrey_big4.png",
            )
        for i in range(STARTING_ASTEROID_COUNT):
            # Pick one of four random rock images
            image_no = random.randrange(4)

            enemy_sprite = AsteroidSprite(
                image_list[image_no],
                scale=SCALE,
                type=ASTERIOD_TYPE_BIG,
            )

            # Set position
            enemy_sprite.center_y = random.randrange(BOTTOM_LIMIT, TOP_LIMIT)
            enemy_sprite.center_x = random.randrange(LEFT_LIMIT, RIGHT_LIMIT)

            # Set speed / rotation
            enemy_sprite.change_x = random.random() * 2 - 1
            enemy_sprite.change_y = random.random() * 2 - 1
            enemy_sprite.change_angle = (random.random() - 0.5) * 2

            self.asteroid_list.append(enemy_sprite)

        self.text_score.text = f"Score: {self.score}"
        self.text_asteroid_count.text = f"Asteroid Count: {len(self.asteroid_list)}"

    def on_draw(self):
        """ Render the screen """

        # Clear the screen before we start drawing
        self.clear()

        # Draw all the sprites.
        self.asteroid_list.draw()
        self.ship_life_list.draw()
        self.bullet_list.draw()
        self.player_sprite_list.draw()

        # Draw the text
        self.text_score.draw()
        self.text_asteroid_count.draw()

    def on_key_press(self, symbol, modifiers):
        """ Called whenever a key is pressed. """
        # Shoot if the player hit the space bar and we aren't respawning.
        if not self.player_sprite.respawning and symbol == arcade.key.SPACE:
            bullet_sprite = TurningSprite(":resources:images/space_shooter/laserBlue01.png",
                                          scale=SCALE)

            # Set bullet vector
            bullet_speed = 13
            angle_radians = math.radians(self.player_sprite.angle)
            bullet_sprite.change_y = math.cos(angle_radians) * bullet_speed
            bullet_sprite.change_x = math.sin(angle_radians) * bullet_speed

            # Set bullet position
            bullet_sprite.center_x = self.player_sprite.center_x
            bullet_sprite.center_y = self.player_sprite.center_y

            # Add to our sprite list
            self.bullet_list.append(bullet_sprite)

            # Go ahead and move it a frame
            bullet_sprite.update()

            # Pew pew
            arcade.play_sound(self.laser_sound, speed=random.random() * 3 + 0.5)

        if symbol == arcade.key.LEFT:
            self.player_sprite.change_angle = -TURN_SPEED
        elif symbol == arcade.key.RIGHT:
            self.player_sprite.change_angle = TURN_SPEED
        elif symbol == arcade.key.UP:
            self.player_sprite.thrust = THRUST_AMOUNT
        elif symbol == arcade.key.DOWN:
            self.player_sprite.thrust = -THRUST_AMOUNT
        # Restart the game if the player hits 'R'
        elif symbol == arcade.key.R:
            self.start_new_game()
        # Quit if the player hits escape
        elif symbol == arcade.key.ESCAPE:
            self.window.close()

    def on_key_release(self, symbol, modifiers):
        """ Called whenever a key is released. """
        if symbol == arcade.key.LEFT:
            self.player_sprite.change_angle = 0
        elif symbol == arcade.key.RIGHT:
            self.player_sprite.change_angle = 0
        elif symbol == arcade.key.UP:
            self.player_sprite.thrust = 0
        elif symbol == arcade.key.DOWN:
            self.player_sprite.thrust = 0

    def split_asteroid(self, asteroid: AsteroidSprite):
        """ Split an asteroid into chunks. """
        x = asteroid.center_x
        y = asteroid.center_y
        self.score += 1

        if asteroid.type == ASTERIOD_TYPE_BIG:
            # Split large asteroid into 2 medium ones
            for i in range(3):
                image_no = random.randrange(2)
                image_list = [":resources:images/space_shooter/meteorGrey_med1.png",
                              ":resources:images/space_shooter/meteorGrey_med2.png"]

                enemy_sprite = AsteroidSprite(image_list[image_no],
                                              scale=SCALE * 1.5,
                                              type=ASTERIOD_TYPE_MEDIUM)

                enemy_sprite.center_y = y
                enemy_sprite.center_x = x

                enemy_sprite.change_x = random.random() * 2.5 - 1.25
                enemy_sprite.change_y = random.random() * 2.5 - 1.25

                enemy_sprite.change_angle = (random.random() - 0.5) * 2

                self.asteroid_list.append(enemy_sprite)
                self.hit_sound1.play()

        elif asteroid.type == ASTERIOD_TYPE_MEDIUM:
            # Split medium asteroid into 2 small ones
            for i in range(3):
                image_no = random.randrange(2)
                image_list = [":resources:images/space_shooter/meteorGrey_small1.png",
                              ":resources:images/space_shooter/meteorGrey_small2.png"]

                enemy_sprite = AsteroidSprite(image_list[image_no],
                                              scale=SCALE * 1.5,
                                              type=ASTERIOD_TYPE_SMALL)

                enemy_sprite.center_y = y
                enemy_sprite.center_x = x

                enemy_sprite.change_x = random.random() * 3 - 1.5
                enemy_sprite.change_y = random.random() * 3 - 1.5

                enemy_sprite.change_angle = (random.random() - 0.5) * 2

                self.asteroid_list.append(enemy_sprite)
                self.hit_sound2.play()

        elif asteroid.type == ASTERIOD_TYPE_SMALL:
            # Split small asteroid into 2 tiny ones
            for i in range(3):
                image_no = random.randrange(2)
                image_list = [":resources:images/space_shooter/meteorGrey_tiny1.png",
                              ":resources:images/space_shooter/meteorGrey_tiny2.png"]

                enemy_sprite = AsteroidSprite(image_list[image_no],
                                              scale=SCALE * 1.5,
                                              type=ASTERIOD_TYPE_TINY)

                enemy_sprite.center_y = y
                enemy_sprite.center_x = x

                enemy_sprite.change_x = random.random() * 3.5 - 1.75
                enemy_sprite.change_y = random.random() * 3.5 - 1.75

                enemy_sprite.change_angle = (random.random() - 0.5) * 2

                self.asteroid_list.append(enemy_sprite)
                self.hit_sound3.play()

        elif asteroid.type == ASTERIOD_TYPE_TINY:
            # Do nothing. The tiny asteroid just goes away.
            self.hit_sound4.play()

    def on_update(self, x):
        """ Move everything """

        if not self.game_over:
            self.asteroid_list.update()
            self.bullet_list.update()
            self.player_sprite_list.update()

            for bullet in self.bullet_list:
                asteroids = arcade.check_for_collision_with_list(bullet,
                                                                 self.asteroid_list)

                for asteroid in asteroids:
                    # expected AsteroidSprite, got Sprite instead
                    self.split_asteroid(cast(AsteroidSprite, asteroid))
                    asteroid.remove_from_sprite_lists()
                    bullet.remove_from_sprite_lists()

                # Remove bullet if it goes off-screen
                size = max(bullet.width, bullet.height)
                if bullet.center_x < 0 - size:
                    bullet.remove_from_sprite_lists()
                if bullet.center_x > WINDOW_WIDTH + size:
                    bullet.remove_from_sprite_lists()
                if bullet.center_y < 0 - size:
                    bullet.remove_from_sprite_lists()
                if bullet.center_y > WINDOW_HEIGHT + size:
                    bullet.remove_from_sprite_lists()

            if not self.player_sprite.respawning:
                asteroids = arcade.check_for_collision_with_list(self.player_sprite,
                                                                 self.asteroid_list)
                if len(asteroids) > 0:
                    if self.lives > 0:
                        self.lives -= 1
                        self.player_sprite.respawn()
                        self.split_asteroid(cast(AsteroidSprite, asteroids[0]))
                        asteroids[0].remove_from_sprite_lists()
                        self.ship_life_list.pop().remove_from_sprite_lists()
                        print("Crash")
                    else:
                        self.game_over = True
                        print("Game over")

        # Update the text objects
        self.text_score.text = f"Score: {self.score}"
        self.text_asteroid_count.text = f"Asteroid Count: {len(self.asteroid_list)}"


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.start_new_game()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Platformer Game

python -m arcade.examples.platform_tutorial.03_more_sprites
"""
import arcade

# Constants
WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Platformer"

# Constants used to scale our sprites from their original size
TILE_SCALING = 0.5


class GameView(arcade.Window):
    """
    Main application class.
    """

    def __init__(self):

        # Call the parent class and set up the window
        super().__init__(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

        # Variable to hold our texture for our player
        self.player_texture = arcade.load_texture(
            ":resources:images/animated_characters/female_adventurer/femaleAdventurer_idle.png"
        )

        # Separate variable that holds the player sprite
        self.player_sprite = arcade.Sprite(self.player_texture)
        self.player_sprite.center_x = 64
        self.player_sprite.center_y = 128

        # SpriteList for our player
        self.player_list = arcade.SpriteList()
        self.player_list.append(self.player_sprite)

        # SpriteList for our boxes and ground
        # Putting our ground and box Sprites in the same SpriteList
        # will make it easier to perform collision detection against
        # them later on. Setting the spatial hash to True will make
        # collision detection much faster if the objects in this
        # SpriteList do not move.
        self.wall_list = arcade.SpriteList(use_spatial_hash=True)

        # Create the ground
        # This shows using a loop to place multiple sprites horizontally
        for x in range(0, 1250, 64):
            wall = arcade.Sprite(":resources:images/tiles/grassMid.png", scale=0.5)
            wall.center_x = x
            wall.center_y = 32
            self.wall_list.append(wall)

        # Put some crates on the ground
        # This shows using a coordinate list to place sprites
        coordinate_list = [[512, 96], [256, 96], [768, 96]]

        for coordinate in coordinate_list:
            # Add a crate on the ground
            wall = arcade.Sprite(
                ":resources:images/tiles/boxCrate_double.png", scale=0.5
            )
            wall.position = coordinate
            self.wall_list.append(wall)

        self.background_color = arcade.csscolor.CORNFLOWER_BLUE

    def setup(self):
        """Set up the game here. Call this function to restart the game."""
        pass

    def on_draw(self):
        """Render the screen."""

        # Clear the screen to the background color
        self.clear()

        # Draw our sprites
        self.player_list.draw()
        self.wall_list.draw()


def main():
    """Main function"""
    window = GameView()
    window.setup()
    arcade.run()


if __name__ == "__main__":
    main()



"""
This program shows how to have a pause screen without resetting the game.

Make a separate class for each view (screen) in your game.
The class will inherit from arcade.View. The structure will
look like an arcade.Window as each View will need to have its own draw,
update and window event methods. To switch a View, simply create a view
with `view = MyView()` and then use the "self.window.set_view(view)" method.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.view_pause_screen
"""

import arcade

WIDTH = 1280
HEIGHT = 720
SPRITE_SCALING = 0.5


class MenuView(arcade.View):
    def on_show_view(self):
        self.window.background_color = arcade.color.WHITE

    def on_draw(self):
        self.clear()
        arcade.draw_text("Menu Screen", WIDTH / 2, HEIGHT / 2,
                         arcade.color.BLACK, font_size=50, anchor_x="center")
        arcade.draw_text("Click to advance.", WIDTH / 2, HEIGHT / 2 - 75,
                         arcade.color.GRAY, font_size=20, anchor_x="center")

    def on_mouse_press(self, _x, _y, _button, _modifiers):
        game = GameView()
        self.window.show_view(game)


class GameView(arcade.View):
    def __init__(self):
        super().__init__()
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=SPRITE_SCALING,
        )
        self.player_sprite.center_x = 50
        self.player_sprite.center_y = 50
        self.player_sprite.velocity = 3, 3

    def on_show_view(self):
        self.window.background_color = arcade.color.AMAZON

    def on_draw(self):
        self.clear()
        # Draw all the sprites.
        arcade.draw_sprite(self.player_sprite)

        # Show tip to pause screen
        arcade.draw_text("Press Esc. to pause",
                         WIDTH // 2,
                         HEIGHT - 100,
                         arcade.color.BLACK,
                         font_size=20,
                         anchor_x="center")

    def on_update(self, delta_time):
        # Call update on all sprites
        self.player_sprite.update()

        # Bounce off the edges
        if self.player_sprite.left < 0 or self.player_sprite.right > WIDTH:
            self.player_sprite.change_x *= -1
        if self.player_sprite.bottom < 0 or self.player_sprite.top > HEIGHT:
            self.player_sprite.change_y *= -1

    def on_key_press(self, key, _modifiers):
        if key == arcade.key.ESCAPE:
            # pass self, the current view, to preserve this view's state
            pause = PauseView(self)
            self.window.show_view(pause)


class PauseView(arcade.View):
    def __init__(self, game_view):
        super().__init__()
        self.game_view = game_view

    def on_show_view(self):
        self.window.background_color = arcade.color.ORANGE

    def on_draw(self):
        self.clear()

        # Draw player, for effect, on pause screen.
        # The previous View (GameView) was passed in
        # and saved in self.game_view.
        player_sprite = self.game_view.player_sprite
        arcade.draw_sprite(player_sprite)

        # draw an orange filter over him
        arcade.draw_lrbt_rectangle_filled(left=player_sprite.left,
                                          right=player_sprite.right,
                                          bottom=player_sprite.bottom,
                                          top=player_sprite.top,
                                          color=arcade.color.ORANGE[:3] + (200,))

        arcade.draw_text("PAUSED", WIDTH / 2, HEIGHT / 2 + 50,
                         arcade.color.BLACK, font_size=50, anchor_x="center")

        # Show tip to return or reset
        arcade.draw_text("Press Esc. to return",
                         WIDTH / 2,
                         HEIGHT / 2,
                         arcade.color.BLACK,
                         font_size=20,
                         anchor_x="center")
        arcade.draw_text("Press Enter to reset",
                         WIDTH / 2,
                         HEIGHT / 2 - 30,
                         arcade.color.BLACK,
                         font_size=20,
                         anchor_x="center")

    def on_key_press(self, key, _modifiers):
        if key == arcade.key.ESCAPE:   # resume game
            self.window.show_view(self.game_view)
        elif key == arcade.key.ENTER:  # reset game
            game = GameView()
            self.window.show_view(game)


def main():
    window = arcade.Window(WIDTH, HEIGHT, "Instruction and Game Over Views Example")
    menu = MenuView()
    window.show_view(menu)
    arcade.run()


if __name__ == "__main__":
    main()



"""
Move with a Sprite Animation

Simple program to show basic sprite usage.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_move_animation
"""
import arcade
import random

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Move with a Sprite Animation Example"

COIN_SCALE = 0.5
COIN_COUNT = 50
CHARACTER_SCALING = 1

# How fast to move, and how fast to run the animation
MOVEMENT_SPEED = 5
UPDATES_PER_FRAME = 5

# Constants used to track if the player is facing left or right
RIGHT_FACING = 0
LEFT_FACING = 1


class PlayerCharacter(arcade.Sprite):
    def __init__(self, idle_texture_pair, walk_texture_pairs):
        # Default to face-right
        self.character_face_direction = RIGHT_FACING

        # Used for flipping between image sequences
        self.cur_texture = 0
        self.idle_texture_pair = idle_texture_pair
        self.walk_textures = walk_texture_pairs

        # Adjust the collision box. Default includes too much empty space
        # side-to-side. Box is centered at sprite center, (0, 0)
        self.points = [[-22, -64], [22, -64], [22, 28], [-22, 28]]
        # Set up parent class
        super().__init__(self.idle_texture_pair[0], scale=CHARACTER_SCALING)


    def update_animation(self, delta_time: float = 1 / 60):

        # Figure out if we need to flip face left or right
        if self.change_x < 0 and self.character_face_direction == RIGHT_FACING:
            self.character_face_direction = LEFT_FACING
        elif self.change_x > 0 and self.character_face_direction == LEFT_FACING:
            self.character_face_direction = RIGHT_FACING

        # Idle animation
        if self.change_x == 0 and self.change_y == 0:
            self.texture = self.idle_texture_pair[self.character_face_direction]
            return

        # Walking animation
        self.cur_texture += 1
        if self.cur_texture > 7 * UPDATES_PER_FRAME:
            self.cur_texture = 0
        frame = self.cur_texture // UPDATES_PER_FRAME
        direction = self.character_face_direction
        self.texture = self.walk_textures[frame][direction]


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        """ Set up the game and initialize the variables. """
        super().__init__()

        # Sprite lists
        self.player_list = None
        self.coin_list = None

        # Set up the player
        self.score = 0
        self.score_text = arcade.Text("Score: 0", 10, 20, arcade.color.WHITE, 14)
        self.player = None

        # --- Load Textures for the player ---
        # Images from Kenney.nl's Asset Pack 3. We pick one randomly
        character_types = [
            ":resources:images/animated_characters/female_adventurer/femaleAdventurer",
            ":resources:images/animated_characters/female_person/femalePerson",
            ":resources:images/animated_characters/male_person/malePerson",
            ":resources:images/animated_characters/male_adventurer/maleAdventurer",
            ":resources:images/animated_characters/zombie/zombie",
            ":resources:images/animated_characters/robot/robot",
        ]
        chosen_character = random.choice(character_types)

        # Load textures for idle standing
        idle_texture = arcade.load_texture(f"{chosen_character}_idle.png")
        self.idle_texture_pair = idle_texture, idle_texture.flip_left_right()

        # Load textures for walking
        self.walk_texture_pairs = []
        for i in range(8):
            texture = arcade.load_texture(f"{chosen_character}_walk{i}.png")
            self.walk_texture_pairs.append((texture, texture.flip_left_right()))

    def setup(self):
        self.player_list = arcade.SpriteList()
        self.coin_list = arcade.SpriteList()

        # Set up the player
        self.score = 0
        self.player = PlayerCharacter(self.idle_texture_pair, self.walk_texture_pairs)
        self.player.position = self.center
        self.player.scale = 0.8

        self.player_list.append(self.player)

        for i in range(COIN_COUNT):
            coin = arcade.Sprite(":resources:images/items/gold_1.png", scale=0.5)
            coin.center_x = random.randrange(WINDOW_WIDTH)
            coin.center_y = random.randrange(WINDOW_HEIGHT)

            self.coin_list.append(coin)

        # Set the background color
        self.background_color = arcade.color.AMAZON

    def on_draw(self):
        """
        Render the screen.
        """

        # This command has to happen before we start drawing
        self.clear()

        # Draw all the sprites.
        self.coin_list.draw()
        self.player_list.draw()

        # Put the text on the screen.
        self.score_text.text = f"Score: {self.score}"
        self.score_text.draw()

    def on_key_press(self, key, modifiers):
        """
        Called whenever a key is pressed.
        """
        # Player controls for movement using arrow keys and WASD
        if key in (arcade.key.UP, arcade.key.W):
            self.player.change_y = MOVEMENT_SPEED
        elif key in (arcade.key.DOWN, arcade.key.S):
            self.player.change_y = -MOVEMENT_SPEED
        elif key in (arcade.key.LEFT, arcade.key.A):
            self.player.change_x = -MOVEMENT_SPEED
        elif key in (arcade.key.RIGHT, arcade.key.D):
            self.player.change_x = MOVEMENT_SPEED
        # Quit
        elif key in (arcade.key.ESCAPE, arcade.key.Q):
            arcade.close_window()

    def on_key_release(self, key, modifiers):
        """
        Called when the user releases a key.
        """
        if key in (arcade.key.UP, arcade.key.DOWN, arcade.key.W, arcade.key.S):
            self.player.change_y = 0
        elif key in (arcade.key.LEFT, arcade.key.RIGHT, arcade.key.A, arcade.key.D):
            self.player.change_x = 0

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Move the player
        self.player_list.update()

        # Update the players animation
        self.player_list.update_animation()

        # Generate a list of all sprites that collided with the player.
        hit_list = arcade.check_for_collision_with_list(self.player, self.coin_list)

        # Loop through each colliding sprite, remove it, and add to the score.
        for coin in hit_list:
            coin.remove_from_sprite_lists()
            self.score += 1


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from __future__ import annotations

from weakref import WeakValueDictionary

import PIL
import PIL.Image

import arcade
import arcade.cache
from arcade.texture import (
    ImageData,
    Texture,
    make_circle_texture,
    make_soft_circle_texture,
)
from arcade.types import RGBA255, Color
from arcade.types.rect import Rect

from .sprite import Sprite


class SpriteSolidColor(Sprite):
    """
    A rectangular sprite of the given ``width``, ``height``, and ``color``.

    The texture is automatically generated instead of loaded from a
    file. Internally only a single global image is used for this
    sprite type so concerns about memory usage non-existent regardless
    of size or number of sprite variations.

    Different texture configurations (width, height) are weakly cached internally
    to avoid creating multiple textures with the same configuration.

    Args:
        width:
            Width of the sprite in pixels
        height:
            Height of the sprite in pixels
        center_x:
            Initial x position of the sprite
        center_y:
            Initial y position of the sprite
        color:
            The color of the sprite as a :py:class:`~arcade.types.Color`,
            an RGBA tuple, or an RGB tuple.
        angle:
            Initial angle of the sprite in degrees
    """

    __slots__ = ()
    _default_image: ImageData | None = None
    # To avoid making lots of texture instances with the same configuration
    # we cache them here weakly. Making a 100 x 100 grid of white sprites
    # only create 1 texture instead of 1000. This saves memory and processing
    # time for the default texture atlas.
    _texture_cache: WeakValueDictionary[tuple[int, int], Texture] = WeakValueDictionary()

    def __init__(
        self,
        width: int,
        height: int,
        center_x: float = 0,
        center_y: float = 0,
        color: RGBA255 = Color(255, 255, 255, 255),
        angle: float = 0,
        **kwargs,
    ):
        texture = self.__class__._texture_cache.get((width, height))
        if texture is None:
            texture = Texture(
                self._get_default_image(),
                hit_box_points=(
                    (-width / 2, -height / 2),
                    (width / 2, -height / 2),
                    (width / 2, height / 2),
                    (-width / 2, height / 2),
                ),
            )
            texture.size = width, height
            self.__class__._texture_cache[(width, height)] = texture

        super().__init__(
            texture,
            center_x=center_x,
            center_y=center_y,
            angle=angle,
        )
        self.color = Color.from_iterable(color)

    @classmethod
    def from_rect(cls, rect: Rect, color: Color, angle: float = 0.0) -> SpriteSolidColor:
        """
        Construct a new SpriteSolidColor from a :py:class:`~arcade.types.rect.Rect`.

        Args:
            rect:
                The rectangle to use for the sprite's dimensions and position.
            color:
                The color of the sprite as a :py:class:`~arcade.types.Color`,
                an RGBA tuple, or an RGB tuple.
            angle:
                The angle of the sprite in degrees.
        """
        return cls(int(rect.width), int(rect.height), rect.x, rect.y, color, angle)

    def _get_default_image(self) -> ImageData:
        """Lazy-load the default image for this sprite type."""
        im = self.__class__._default_image
        if im is None:
            im = ImageData(
                PIL.Image.new("RGBA", size=(32, 32), color=(255, 255, 255, 255)),
                hash="sprite_solid_color",
            )
            self.__class__._default_image = im
        return im


class SpriteCircle(Sprite):
    """
    A circle of the specified `radius <https://simple.wikipedia.org/wiki/Radius>`_.

    The texture is automatically generated instead of loaded from a
    file.

    There may be a stutter the first time a combination of ``radius``,
    ``color``, and ``soft`` is used due to texture generation. All
    subsequent calls for the same combination will run faster because
    they will re-use the texture generated earlier.

    For a gradient fill instead of a solid color, set ``soft`` to
    ``True``. The circle will fade from an opaque center to transparent
    at the edges.

    Args:
        radius:
            Radius of the circle in pixels
        color:
            The Color of the sprite as an RGB or RGBA tuple
        soft:
            If ``True``, the circle will fade from an opaque
            center to transparent edges.
    """

    # Local weak cache for textures to avoid creating multiple instances with the same configuration
    _texture_cache: WeakValueDictionary[tuple[int, RGBA255, bool], Texture] = WeakValueDictionary()

    def __init__(self, radius: int, color: RGBA255, soft: bool = False, **kwargs):
        radius = int(radius)
        diameter = radius * 2

        cache_key = diameter, color, soft
        # Get existing texture from cache if possible
        texture = self.__class__._texture_cache.get(cache_key)

        if not texture:
            if soft:
                texture = make_soft_circle_texture(
                    diameter,
                    color=(255, 255, 255, 255),
                    hit_box_algorithm=arcade.hitbox.algo_simple,
                )
            else:
                texture = make_circle_texture(
                    diameter,
                    color=(255, 255, 255, 255),
                )

            self.__class__._texture_cache[cache_key] = texture

        # apply results to the new sprite
        super().__init__(texture)
        self.color = Color.from_iterable(color)



from pathlib import Path

import PIL.Image

from arcade.texture import Texture
from arcade.resources import resolve
from .base import BasicSprite, SpriteType, SpriteType_co
from .sprite import Sprite
from .mixins import PymunkMixin, PyMunk
from .animated import (
    TextureAnimationSprite,
    TextureAnimation,
    TextureKeyframe,
    AnimatedWalkingSprite,
)
from .colored import SpriteSolidColor, SpriteCircle
from .enums import (
    FACE_LEFT,
    FACE_RIGHT,
    FACE_UP,
    FACE_DOWN,
)


def load_animated_gif(resource_name: str | Path) -> TextureAnimationSprite:
    """
    Attempt to load an animated GIF as a :class:`TextureAnimationSprite`.

    .. note::

        Many older GIFs will load with incorrect transparency for every
        frame but the first. Until the Pillow library handles the quirks of
        the format better, loading animated GIFs will be pretty buggy. A
        good workaround is loading GIFs in another program and exporting them
        as PNGs, either as sprite sheets or a frame per file.

    Args:
        resource_name: A path to a GIF as either a :py:class:`pathlib.Path`
            or a :py:class:`str` which may include a
            :ref:`resource handle <resource_handles>`.

    """

    file_name = resolve(resource_name)
    image_object = PIL.Image.open(file_name)

    # Pillow doc recommends testing for the is_animated attribute as of 10.0.0
    # https://pillow.readthedocs.io/en/stable/deprecations.html#categories
    if not getattr(image_object, "is_animated", False) or not (
        n_frames := getattr(image_object, "n_frames", 0)
    ):
        raise TypeError(f"The file {resource_name} is not an animated gif.")

    sprite = TextureAnimationSprite()
    keyframes = []
    for frame in range(n_frames):
        image_object.seek(frame)
        frame_duration = image_object.info["duration"]
        image = image_object.convert("RGBA")
        texture = Texture(image)
        texture.file_path = file_name
        # sprite.textures.append(texture)
        keyframes.append(TextureKeyframe(texture, frame_duration))

    animation = TextureAnimation(keyframes=keyframes)
    sprite.animation = animation
    return sprite


__all__ = [
    "SpriteType",
    "SpriteType_co",
    "BasicSprite",
    "Sprite",
    "PyMunk",
    "TextureAnimationSprite",
    "TextureAnimation",
    "TextureKeyframe",
    "AnimatedWalkingSprite",
    "load_animated_gif",
    "SpriteSolidColor",
    "SpriteCircle",
    "FACE_LEFT",
    "FACE_RIGHT",
    "FACE_UP",
    "FACE_DOWN",
    "PymunkMixin",
]



from pyglet.math import Vec3

from arcade.camera import CameraData
from arcade.math import quaternion_rotation
from arcade.types import Point3


def look_at(camera: CameraData, target: Point3, up: Point3 | None = None) -> tuple[Point3, Point3]:
    """
    Calculate the neccisary forward and up vectors to have the camera look towards
    the target point.

    Uses the camera's up if none is provided.
    Args:
        camera: The CameraData to work from
        target: The point in 3D world space to look at
        up: An optional up axis to refer to when calculating the true up vector.
    """
    px, py, pz = camera.position
    tx, ty, tz = target
    dx, dy, dz = tx - px, ty - py, tz - pz

    up = up or camera.up

    f = Vec3(dx, dy, dz).normalize()
    r = f.cross(up)
    u = r.cross(f).normalize()

    return f, u


def orbit(camera: CameraData, origin: Point3, axis: Point3, angle: float) -> Point3:
    """
    Find the new position for the camera when rotated around the origin

    Args:
        camera: The CameraData to work from
        origin: The point around which to orbit
        axis: The axis to rotate around, like the stick on a spinning top.
        angle: The angle in degrees to rotate by
    """
    px, py, pz = camera.position
    tx, ty, tz = origin

    rx, ry, rz = quaternion_rotation(axis, (px - tx, py - ty, pz - tz), angle)
    return tx + rx, ty + ry, tz + rz



"""
City Scape Generator

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.shape_list_skylines
"""

import random
import time

import arcade
from arcade.gl import geometry

SCREEN_WIDTH = 1280
SCREEN_HEIGHT = 720
SCREEN_TITLE = "Skyline Using Buffered Shapes"


def make_star_field(star_count):
    """Make a bunch of circles for stars."""

    shape_list = arcade.shape_list.ShapeElementList()

    for star_no in range(star_count):
        x = random.randrange(SCREEN_WIDTH)
        y = random.randrange(SCREEN_HEIGHT)
        radius = random.randrange(1, 4)
        brightness = random.randrange(127, 256)
        color = (brightness, brightness, brightness)
        shape = arcade.shape_list.create_rectangle_filled(x, y, radius, radius, color)
        shape_list.append(shape)

    return shape_list


def make_skyline(
    width,
    skyline_height,
    skyline_color,
    gap_chance=0.70,
    window_chance=0.30,
    light_on_chance=0.5,
    window_color=(255, 255, 200),
    window_margin=3,
    window_gap=2,
    cap_chance=0.20,
):
    """Make a skyline"""

    shape_list = arcade.shape_list.ShapeElementList()

    # Add the "base" that we build the buildings on
    shape = arcade.shape_list.create_rectangle_filled(
        width / 2, skyline_height / 2, width, skyline_height, skyline_color
    )
    shape_list.append(shape)

    building_center_x = 0

    skyline_point_list = []
    color_list = []

    while building_center_x < width:
        # Is there a gap between the buildings?
        if random.random() < gap_chance:
            gap_width = random.randrange(10, 50)
        else:
            gap_width = 0

        # Figure out location and size of building
        building_width = random.randrange(20, 70)
        building_height = random.randrange(40, 150)
        building_center_x += gap_width + (building_width / 2)
        building_center_y = skyline_height + (building_height / 2)

        x1 = building_center_x - building_width / 2
        x2 = building_center_x + building_width / 2
        y1 = skyline_height
        y2 = skyline_height + building_height

        skyline_point_list.append([x1, y1])

        skyline_point_list.append([x1, y2])

        skyline_point_list.append([x2, y2])

        skyline_point_list.append([x2, y1])

        for i in range(4):
            color_list.append([skyline_color[0], skyline_color[1], skyline_color[2]])

        if random.random() < cap_chance:
            x1 = building_center_x - building_width / 2
            x2 = building_center_x + building_width / 2
            x3 = building_center_x

            y1 = y2 = building_center_y + building_height / 2
            y3 = y1 + building_width / 2

            shape = arcade.shape_list.create_polygon([[x1, y1], [x2, y2], [x3, y3]], skyline_color)
            shape_list.append(shape)

        # See if we should have some windows
        if random.random() < window_chance:
            # Yes windows! How many windows?
            window_rows = random.randrange(10, 15)
            window_columns = random.randrange(1, 7)

            # Based on that, how big should they be?
            window_height = (building_height - window_margin * 2) / window_rows
            window_width = (
                building_width - window_margin * 2 - window_gap * (window_columns - 1)
            ) / window_columns

            # Find the bottom left of the building so we can start adding widows
            building_base_y = building_center_y - building_height / 2
            building_left_x = building_center_x - building_width / 2

            # Loop through each window
            for row in range(window_rows):
                for column in range(window_columns):
                    if random.random() < light_on_chance:
                        x1 = building_left_x + column * (window_width + window_gap) + window_margin
                        x2 = (
                            building_left_x
                            + column * (window_width + window_gap)
                            + window_width
                            + window_margin
                        )
                        y1 = building_base_y + row * window_height
                        y2 = building_base_y + row * window_height + window_height * 0.8

                        skyline_point_list.append([x1, y1])
                        skyline_point_list.append([x1, y2])
                        skyline_point_list.append([x2, y2])
                        skyline_point_list.append([x2, y1])

                        for i in range(4):
                            color_list.append((window_color[0], window_color[1], window_color[2]))

        building_center_x += building_width / 2

    shape = arcade.shape_list.create_rectangles_filled_with_colors(skyline_point_list, color_list)
    shape_list.append(shape)

    return shape_list


class MyGame(arcade.Window):
    """Main application class."""

    def __init__(self):
        """Initializer"""
        # Call the parent class initializer
        super().__init__(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_TITLE, antialiasing=True)

        self.stars = make_star_field(150)
        self.skyline1 = make_skyline(SCREEN_WIDTH * 5, 250, (80, 80, 80))
        self.skyline2 = make_skyline(SCREEN_WIDTH * 5, 150, (50, 50, 50))

        arcade.set_background_color(arcade.color.BLACK)

    def setup(self):
        """Set up the game and initialize the variables."""
        # Offscreen stuff
        self.program = self.ctx.program(
            vertex_shader="""
                #version 330

                in vec2 in_vert;
                in vec2 in_uv;
                out vec2 v_uv;

                void main() {
                    gl_Position = vec4(in_vert, 0.0, 1.0);
                    v_uv = in_uv;
                }
            """,
            fragment_shader="""
                #version 330

                uniform sampler2D tex;
                uniform float angle;
                uniform float zoom;

                in vec2 v_uv;
                out vec4 f_color;

                void main() {
                    mat3 rotate = mat3(
                        cos(angle), -sin(angle), 0.5,
                        sin(angle),  cos(angle), 0.5,
                        0.0,         0.0,        1.0
                    );
                    mat3 scale = mat3(
                         1.0,  0.0,  0.0,
                         0.0,  0.5,  0.0,
                         0.0,  0.0,  1.0
                    );
                    f_color = texture(
                        tex, (rotate * scale * vec3(v_uv + vec2(-0.5, -0.5), 1.0)).xy * zoom
                    );
                }
            """,
        )
        self.color_attachment = self.ctx.texture((SCREEN_WIDTH, SCREEN_HEIGHT))
        self.offscreen = self.ctx.framebuffer(color_attachments=[self.color_attachment])
        self.quad_fs = geometry.quad_2d_fs()
        self.t0 = time.time()
        self.frame = 0

    def on_draw(self):
        """
        Render the screen.
        """

        start_time = int(round(time.time() * 1000))
        self.clear()

        self.offscreen.use()
        self.offscreen.clear()
        self.stars.draw()
        self.skyline1.draw()
        self.skyline2.draw()

        self.use()
        self.color_attachment.use(0)
        self.program["angle"] = -1 + (time.time() - self.t0) / 5
        self.program["zoom"] = 3 + (time.time() - self.t0) / 5
        self.quad_fs.render(self.program)

        end_time = int(round(time.time() * 1000))
        total_time = end_time - start_time

        arcade.draw_text(f"Time: {total_time}", 10, 10, arcade.color.WHITE)

        # # Dumping gif anim
        # from arcade import get_image
        # image = get_image()
        # image.save(f"test/frame_{str(self.frame).zfill(4)}.png", 'PNG')
        # self.frame += 1

    def on_update(self, delta_time):
        """Movement and game logic"""
        self.skyline1.center_x -= 0.5
        self.skyline2.center_x -= 1


def main():
    window = MyGame()
    window.setup()
    arcade.run()


if __name__ == "__main__":
    main()



"""
A scrolling Background.

This program loads a texture from a file,
and create a screen sized background.
The background is constantly aligned to the screen, and the
texture offset changed. This creates an illusion of moving.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.background_scrolling
"""
import arcade
import arcade.future.background as background

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720

WINDOW_TITLE = "Scrolling Background Example"

PLAYER_SPEED = 300
CAMERA_SPEED = 0.5


class GameView(arcade.View):
    def __init__(self):
        super().__init__()
        self.camera = arcade.Camera2D()

        # Load the background from file. Sized to match the screen
        self.background = background.Background.from_file(
            ":resources:/images/tiles/sandCenter.png",
            size=(WINDOW_WIDTH, WINDOW_HEIGHT),
        )

        # Create the player sprite.
        self.player_sprite = arcade.SpriteSolidColor(20, 30, color=arcade.color.PURPLE)
        self.player_sprite.center_y = self.camera.viewport_height // 2
        self.player_sprite.center_x = self.camera.viewport_width // 2

        # Track Player Motion
        self.x_direction = 0
        self.y_direction = 0

    def pan_camera_to_player(self):
        # This will center the camera on the player.
        target_x = self.player_sprite.center_x
        target_y = self.player_sprite.center_y
        self.camera.position = arcade.math.lerp_2d(
            self.camera.position,
            (target_x, target_y),
            CAMERA_SPEED,
        )

    def on_update(self, delta_time: float):
        new_position = (
            self.player_sprite.center_x + self.x_direction * delta_time,
            self.player_sprite.center_y + self.y_direction * delta_time,
        )
        self.player_sprite.position = new_position

        self.pan_camera_to_player()

    def on_draw(self):
        self.clear()

        with self.camera.activate():
            # Ensure the background aligns with the camera
            self.background.pos = self.camera.bottom_left

            # Offset the background texture.
            self.background.texture.offset = self.camera.bottom_left

            self.background.draw()
            arcade.draw_sprite(self.player_sprite)

    def on_key_press(self, symbol: int, modifiers: int):
        if symbol in (arcade.key.LEFT, arcade.key.A):
            self.x_direction -= PLAYER_SPEED
        elif symbol in (arcade.key.RIGHT, arcade.key.D):
            self.x_direction += PLAYER_SPEED
        elif symbol in (arcade.key.DOWN, arcade.key.S):
            self.y_direction -= PLAYER_SPEED
        elif symbol in (arcade.key.UP, arcade.key.W):
            self.y_direction += PLAYER_SPEED
        # Close the window if the user presses the escape key
        elif symbol == arcade.key.ESCAPE:
            self.window.close()

    def on_key_release(self, symbol: int, modifiers: int):
        if symbol in (arcade.key.LEFT, arcade.key.A):
            self.x_direction += PLAYER_SPEED
        elif symbol in (arcade.key.RIGHT, arcade.key.D):
            self.x_direction -= PLAYER_SPEED
        elif symbol in (arcade.key.DOWN, arcade.key.S):
            self.y_direction += PLAYER_SPEED
        elif symbol in (arcade.key.UP, arcade.key.W):
            self.y_direction -= PLAYER_SPEED

    def on_resize(self, width: int, height: int):
        super().on_resize(width, height)
        self.camera.match_window()

        # This is to ensure the background covers the entire screen.
        self.background.size = (width, height)


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Literal

from PIL import Image
from PIL.Image import Transpose

from arcade.resources import resolve

# from arcade import Texture
from arcade.texture import Texture
from arcade.types.rect import Rect

if TYPE_CHECKING:
    from arcade.hitbox import HitBoxAlgorithm

OriginChoices = Literal["upper_left", "lower_left"]


class SpriteSheet:
    """
    A sprite sheet is a single image containing multiple smaller images, or
    frames. The class is used to load the image providing methods to slice out
    parts of the image as separate images or textures.

    Note that the default coordinate system used for slicing is using image coordinates
    (0, 0) in the upper left corner. This matches the coordinate system used by PIL.

    Args:
        path Path to the image to load.
        image: PIL image to use.
    """

    def __init__(
        self,
        path: str | Path | None = None,
        image: Image.Image | None = None,
    ):
        self._path = None
        if path:
            self._path = resolve(path)
            self._image = Image.open(self._path).convert("RGBA")
        elif image:
            self._image = image
        else:
            raise ValueError("Must provide either path or image")

        self._flip_flags = (False, False)

    @classmethod
    def from_image(cls, image: Image.Image) -> SpriteSheet:
        """
        Create a sprite sheet from a PIL image.

        Args:
            image: PIL image to use.
        """
        return cls(image=image)

    @property
    def image(self) -> Image.Image:
        """
        Get or set the PIL image for this sprite sheet.
        """
        return self._image

    @image.setter
    def image(self, image: Image.Image):
        self._image = image

    @property
    def path(self) -> Path | None:
        """
        The path to the sprite sheet if it was loaded from disk.
        """
        return self._path

    @property
    def flip_flags(self) -> tuple[bool, bool]:
        """
        Query the orientation of the sprite sheet.
        This can be used to determine if the sprite sheet needs to be flipped.

        Default values are ``(False, False)``. Will be modified when
        :py:meth:`flip_left_right` or :py:meth:`flip_top_bottom` is called.

        Tuple of booleans ``(flip_left_right, flip_top_bottom)``.
        """
        return self._flip_flags

    def flip_left_right(self) -> None:
        """
        Flips the internal image left to right.
        """
        self._image = self._image.transpose(Transpose.FLIP_LEFT_RIGHT)
        self._flip_flags = (not self._flip_flags[0], self._flip_flags[1])

    def flip_top_bottom(self) -> None:
        """
        Flip the internal image top to bottom.
        """
        self._image = self._image.transpose(Transpose.FLIP_TOP_BOTTOM)
        self._flip_flags = (self._flip_flags[0], not self._flip_flags[1])

    def get_image(self, rect: Rect, y_up=False) -> Image.Image:
        """
        Slice out an image from the sprite sheet.

        Args:
            rect:
                The rectangle to crop out.
            y_up:
                Sets the coordinate space of the image to assert (0, 0)
                in the bottom left.
        """
        # PIL box is a 4-tuple: left, upper, right, and lower
        if y_up:
            return self.image.crop(
                (
                    rect.left,
                    self.image.height - rect.bottom - rect.height,
                    rect.right,
                    self.image.height - rect.bottom,
                )
            )
        else:
            return self.image.crop(
                (
                    rect.left,
                    rect.bottom,
                    rect.right,
                    rect.top,
                )
            )

    # slice an image out of the sprite sheet
    def get_texture(
        self, rect: Rect, hit_box_algorithm: HitBoxAlgorithm | None = None, y_up=False
    ) -> Texture:
        """
        Slice out texture from the sprite sheet.

        Args:
            rect:
                The rectangle to crop out.
            hit_box_algorithm:
                Hit box algorithm to use for the texture.
                If not provided, the default hit box algorithm will be used.
            y_up:
                Sets the coordinate space of the image to assert (0, 0)
                in the bottom left.
        """
        im = self.get_image(rect, y_up)
        texture = Texture(im, hit_box_algorithm=hit_box_algorithm)
        texture.file_path = self._path
        texture.crop_values = rect.lbwh_int
        return texture

    def get_image_grid(
        self,
        size: tuple[int, int],
        columns: int,
        count: int,
        margin: tuple[int, int, int, int] = (0, 0, 0, 0),
    ) -> list[Image.Image]:
        """
        Slice a grid of textures from the sprite sheet.

        Args:
            size:
                Size of each texture ``(width, height)``
            columns:
                Number of columns in the grid
            count:
                Number of textures to crop
            margin:
                The margin around each texture ``(left, right, bottom, top)``
        """
        images = []
        width, height = size
        left, right, bottom, top = margin

        for sprite_no in range(count):
            row = sprite_no // columns
            column = sprite_no % columns

            x = (width + left + right) * column
            y = (height + top + bottom) * row
            im = self.image.crop((x, y, x + width, y + height))
            images.append(im)

        return images

    def get_texture_grid(
        self,
        size: tuple[int, int],
        columns: int,
        count: int,
        margin: tuple[int, int, int, int] = (0, 0, 0, 0),
        hit_box_algorithm: HitBoxAlgorithm | None = None,
    ) -> list[Texture]:
        """
        Slice a grid of textures from the sprite sheet.

        Args:
            size:
                Size of each texture ``(width, height)``
            columns:
                Number of columns in the grid
            count:
                Number of textures to crop
            margin:
                The margin around each texture ``(left, right, bottom, top)``
            hit_box_algorithm:
                Hit box algorithm to use for the textures.
                If not provided, the default hit box algorithm will be used.
        """
        textures = []
        width, height = size
        left, right, bottom, top = margin

        for sprite_no in range(count):
            row = sprite_no // columns
            column = sprite_no % columns

            x = (width + left + right) * column
            y = (height + top + bottom) * row
            im = self.image.crop((x, y, x + width, y + height))

            texture = Texture(im, hit_box_algorithm=hit_box_algorithm)
            texture.file_path = self._path
            texture.crop_values = x, y, width, height
            textures.append(texture)

        return textures



"""
Sprite with Moving Platforms

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_moving_platforms
"""
import arcade

SPRITE_SCALING = 0.5

WINDOW_WIDTH = 1289
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprite with Moving Platforms Example"
SPRITE_PIXEL_SIZE = 128
GRID_PIXEL_SIZE = (SPRITE_PIXEL_SIZE * SPRITE_SCALING)

# How many pixels to keep as a minimum margin between the character
# and the edge of the screen.
VIEWPORT_MARGIN = SPRITE_PIXEL_SIZE * SPRITE_SCALING
RIGHT_MARGIN = 4 * SPRITE_PIXEL_SIZE * SPRITE_SCALING

# Physics
MOVEMENT_SPEED = 10 * SPRITE_SCALING
JUMP_SPEED = 28 * SPRITE_SCALING
GRAVITY = .9 * SPRITE_SCALING

# How fast the camera pans to the player. 1.0 is instant.
CAMERA_SPEED = 0.1


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        """ Initializer """

        # Call the parent init
        super().__init__()

        # Sprite lists

        # Drawing non-moving walls separate from moving walls improves performance.
        self.static_wall_list = None
        self.moving_platform_list = None

        self.player_list = None

        # Set up the player
        self.player_sprite = None
        self.physics_engine = None
        self.game_over = False

        # Create the cameras. One for the GUI, one for the sprites.
        # We scroll the 'sprite world' but not the GUI.
        self.camera_sprites = arcade.Camera2D()
        self.camera_gui = arcade.Camera2D()

        self.left_down = False
        self.right_down = False

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Sprite lists
        self.static_wall_list = arcade.SpriteList()
        self.moving_platform_list = arcade.SpriteList()
        self.player_list = arcade.SpriteList()

        # Set up the player
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=SPRITE_SCALING,
        )
        self.player_sprite.center_x = 2 * GRID_PIXEL_SIZE
        self.player_sprite.center_y = 3 * GRID_PIXEL_SIZE
        self.player_list.append(self.player_sprite)

        # Create floor
        for i in range(50):
            wall = arcade.Sprite(":resources:images/tiles/grassMid.png", scale=SPRITE_SCALING)
            wall.bottom = 0
            wall.center_x = - 1000 + i * GRID_PIXEL_SIZE
            self.static_wall_list.append(wall)

        # Create platform side to side
        wall = arcade.Sprite(":resources:images/tiles/grassMid.png", scale=SPRITE_SCALING)
        wall.center_y = 3 * GRID_PIXEL_SIZE
        wall.center_x = 3 * GRID_PIXEL_SIZE
        wall.boundary_left = 2 * GRID_PIXEL_SIZE
        wall.boundary_right = 5 * GRID_PIXEL_SIZE
        wall.change_x = 2 * SPRITE_SCALING
        self.moving_platform_list.append(wall)

        # Create platform side to side
        wall = arcade.Sprite(":resources:images/tiles/grassMid.png", scale=SPRITE_SCALING)
        wall.center_y = 3 * GRID_PIXEL_SIZE
        wall.center_x = 7 * GRID_PIXEL_SIZE
        wall.boundary_left = 5 * GRID_PIXEL_SIZE
        wall.boundary_right = 9 * GRID_PIXEL_SIZE
        wall.change_x = -2 * SPRITE_SCALING
        self.moving_platform_list.append(wall)

        # Create platform moving up and down
        wall = arcade.Sprite(":resources:images/tiles/grassMid.png", scale=SPRITE_SCALING)
        wall.center_y = 5 * GRID_PIXEL_SIZE
        wall.center_x = 5 * GRID_PIXEL_SIZE
        wall.boundary_top = 8 * GRID_PIXEL_SIZE
        wall.boundary_bottom = 4 * GRID_PIXEL_SIZE
        wall.change_y = 2 * SPRITE_SCALING
        self.moving_platform_list.append(wall)

        # Create platform moving diagonally
        wall = arcade.Sprite(":resources:images/tiles/grassMid.png", scale=SPRITE_SCALING)
        wall.center_y = 5 * GRID_PIXEL_SIZE
        wall.center_x = 8 * GRID_PIXEL_SIZE
        wall.boundary_left = 7 * GRID_PIXEL_SIZE
        wall.boundary_right = 9 * GRID_PIXEL_SIZE
        wall.boundary_top = 8 * GRID_PIXEL_SIZE
        wall.boundary_bottom = 4 * GRID_PIXEL_SIZE
        wall.change_x = 2 * SPRITE_SCALING
        wall.change_y = 2 * SPRITE_SCALING
        self.moving_platform_list.append(wall)

        # Create our physics engine
        self.physics_engine = arcade.PhysicsEnginePlatformer(
            self.player_sprite,
            platforms=self.moving_platform_list,
            walls=self.static_wall_list,
            gravity_constant=GRAVITY
        )

        # Set the background color
        self.background_color = arcade.color.AMAZON

        self.game_over = False

    def on_draw(self):
        """
        Render the screen.
        """

        # This command has to happen before we start drawing
        self.clear()

        # Select the camera we'll use to draw all our sprites
        with self.camera_sprites.activate():
            # Draw the sprites
            self.static_wall_list.draw()
            self.moving_platform_list.draw()
            self.player_list.draw()

        # Update & draw our text to the screen
        with self.camera_gui.activate():
            distance = self.player_sprite.right
            output = f"Distance: {distance}"
            arcade.draw_text(output, 10, 20, arcade.color.WHITE, 14)

    def set_x_speed(self):
        if self.left_down and not self.right_down:
            self.player_sprite.change_x = -MOVEMENT_SPEED
        elif self.right_down and not self.left_down:
            self.player_sprite.change_x = MOVEMENT_SPEED
        else:
            self.player_sprite.change_x = 0

    def on_key_press(self, key, modifiers):
        """ Called whenever the mouse moves. """
        if key == arcade.key.UP:
            if self.physics_engine.can_jump():
                self.player_sprite.change_y = JUMP_SPEED
        elif key == arcade.key.LEFT:
            self.left_down = True
            self.set_x_speed()
        elif key == arcade.key.RIGHT:
            self.right_down = True
            self.set_x_speed()

    def on_key_release(self, key, modifiers):
        """ Called when the user presses a mouse button. """
        if key == arcade.key.LEFT:
            self.left_down = False
            self.set_x_speed()
        elif key == arcade.key.RIGHT:
            self.right_down = False
            self.set_x_speed()

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Call update on all sprites
        self.physics_engine.update()

        # Scroll the screen to the player
        self.scroll_to_player()

    def scroll_to_player(self):
        """
        Scroll the window to the player.

        if CAMERA_SPEED is 1, the camera will immediately move to the desired position.
        Anything between 0 and 1 will have the camera move to the location with a smoother
        pan.
        """

        position = (self.player_sprite.center_x, self.player_sprite.center_y)
        self.camera_sprites.position = arcade.math.lerp_2d(
            self.camera_sprites.position,
            position,
            CAMERA_SPEED,
        )


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Example using the TextureAnimationSprite class to animate a sprite using keyframes.
This sprite type is primarily used internally by tilemaps, but can be used for other
purposes as well.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_animated_keyframes
"""
import arcade


class GameView(arcade.View):

    def __init__(self):
        super().__init__()

        # Load the 8 frames for the walking animation
        path = ":assets:/images/animated_characters/female_adventurer/"
        keyframes = [
            arcade.TextureKeyframe(
                texture=arcade.load_texture(
                    path + f"femaleAdventurer_walk{i}.png"
                ),
                duration=100,
            )
            for i in range(8)
        ]
        anim = arcade.TextureAnimation(keyframes=keyframes)
        self.sprite = arcade.TextureAnimationSprite(
            animation=anim,
            scale=5.0,
            center_x=self.width // 2,
            center_y=self.height // 2,
        )

    def on_draw(self):
        self.clear()
        arcade.draw_sprite(self.sprite, pixelated=True)

    def on_update(self, delta_time: float):
        self.sprite.update_animation(delta_time)


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(1280, 720, "Time based animated sprite")

    # Create and setup the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()




"""
Shapes buffered in ShapeElementList

Show how to use a ShapeElementList to display multiple shapes on-screen.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.shapes_buffered
"""

import random

from pyglet import gl

import arcade
from arcade.experimental import postprocessing

# Do the math to figure out our screen dimensions
SCREEN_WIDTH = 1280
SCREEN_HEIGHT = 720
SCREEN_TITLE = "Buffered Shapes"


class MyGame(arcade.Window):
    """
    Main application class.
    """

    def __init__(self, width, height, title):
        """
        Set up the application.
        """
        super().__init__(width, height, title)

        self.shape_list = arcade.shape_list.ShapeElementList()
        self.shape_list.center_x = SCREEN_WIDTH // 2
        self.shape_list.center_y = SCREEN_HEIGHT // 2
        self.shape_list.angle = 0
        # fmt: off
        point_list = (
            (0, 50),
            (10, 10),
            (50, 0),
            (10, -10),
            (0, -50),
            (-10, -10),
            (-50, 0),
            (-10, 10),
            (0, 50),
        )
        # fmt: on
        colors = [
            getattr(arcade.color, color)
            for color in dir(arcade.color)
            if not color.startswith("__")
        ]
        for i in range(5):
            x = SCREEN_WIDTH // 2 - random.randrange(SCREEN_WIDTH - 50)
            y = SCREEN_HEIGHT // 2 - random.randrange(SCREEN_HEIGHT - 50)
            color = random.choice(colors)
            points = [(px + x, py + y) for px, py in point_list]

            my_line_strip = arcade.shape_list.create_line_strip(points, color, 5)
            self.shape_list.append(my_line_strip)

        # fmt: off
        point_list = (
            (-50, -50),
            (0, 40),
            (50, -50),
        )
        # fmt: on

        for i in range(5):
            x = SCREEN_WIDTH // 2 - random.randrange(SCREEN_WIDTH - 50)
            y = SCREEN_HEIGHT // 2 - random.randrange(SCREEN_HEIGHT - 50)
            points = [(px + x, py + y) for px, py in point_list]
            triangle_filled = arcade.shape_list.create_triangles_filled_with_colors(
                points, random.sample(colors, 3)
            )
            self.shape_list.append(triangle_filled)

        # fmt: off
        point_list = (
            (-50, -70),
            (-50, 70),
            (50, 70),
            (50, -70),
        )
        # fmt: on

        for i in range(5):
            x = SCREEN_WIDTH // 2 - random.randrange(SCREEN_WIDTH - 50)
            y = SCREEN_HEIGHT // 2 - random.randrange(SCREEN_HEIGHT - 50)
            points = [(px + x, py + y) for px, py in point_list]
            rect_filled = arcade.shape_list.create_rectangle_filled_with_colors(
                points, random.sample(colors, 4)
            )
            self.shape_list.append(rect_filled)

        # fmt: off
        point_list = (
            (100, 100),
            (50, 150),
            (100, 200),
            (200, 200),
            (250, 150),
            (200, 100),
        )
        # fmt: on
        poly = arcade.shape_list.create_polygon(point_list, (255, 10, 10))
        self.shape_list.append(poly)

        ellipse = arcade.shape_list.create_ellipse(20, 30, 50, 20, (230, 230, 0))
        self.shape_list.append(ellipse)

        self.background_color = arcade.color.BLACK

        self.offscreen = self.ctx.framebuffer(
            color_attachments=self.ctx.texture(
                (SCREEN_WIDTH, SCREEN_HEIGHT),
                wrap_x=gl.GL_CLAMP_TO_EDGE,
                wrap_y=gl.GL_CLAMP_TO_EDGE,
            )
        )
        self.glow = postprocessing.BloomEffect((SCREEN_WIDTH, SCREEN_HEIGHT))

    def on_draw(self):
        """
        Render the screen.
        """
        self.clear()

        self.offscreen.use()
        self.offscreen.clear()
        self.shape_list.draw()

        # This command has to happen before we start drawing
        self.use()
        self.clear()

        gl.glDisable(gl.GL_BLEND)
        self.glow.render(self.offscreen.color_attachments[0], self)
        gl.glEnable(gl.GL_BLEND)

    def on_update(self, delta_time):
        self.shape_list.angle += 0.2
        self.shape_list.center_x += 0.1
        self.shape_list.center_y += 0.1


def main():
    MyGame(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_TITLE)
    arcade.run()


if __name__ == "__main__":
    main()
